# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2012-2017 Guillaume Sagnol
# Copyright (C) 2017-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`Problem`."""

from __future__ import division, print_function

import copy as pycopy
import re
import time
from collections import OrderedDict
from textwrap import TextWrapper

import cvxopt as cvx
import numpy as np

from .. import constraints, expressions, glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_property
from ..compat import MappingProxyType, lru_cache
from ..expressions.data import cvx2np
from ..formatting import natsorted, parameterized_string, picos_box
from ..legacy import deprecated, map_legacy_options, throw_deprecation_warning
from ..solvers import get_solver
from .file_out import write
from .footprint import Footprint
from .options import Options
from .solution import SS_OPTIMAL, Solution

_API_START = api_start(globals())
# -------------------------------


class SolutionFailure(RuntimeError):
    """Solving the problem failed."""

    def __init__(self, code, message):
        """Construct a :exc:`SolutionFailure`.

        :param int code:
            Status code, as defined in :meth:`Problem.solve`.

        :param str message:
            Text description of the failure.
        """
        #: Status code, as defined in :meth:`Problem.solve`.
        self.code = code

        #: Text description of the failure.
        self.message = message

    def __str__(self):
        return "Code {}: {}".format(self.code, self.message)


class Objective:
    """An optimization objective composed of search direction and function."""

    #: Short string denoting a feasibility problem.
    FIND = "find"

    #: Short string denoting a minimization problem.
    MIN  = "min"

    #: Short string denoting a maximization problem.
    MAX  = "max"

    def __init__(self, direction=None, function=None):
        """Construct an optimization objective.

        :param str direction:
            Case insensitive search direction string. One of

            - ``"min"`` or ``"minimize"``,
            - ``"max"`` or ``"maximize"``,
            - ``"find"`` or :obj:`None` (for a feasibility problem).

        :param ~picos.expressions.Expression function:
            The objective function. Must be :obj:`None` for a feasibility
            problem.
        """
        if direction is None:
            direction = self.FIND
        else:
            if not isinstance(direction, str):
                raise TypeError("Search direction must be given as a string.")

            lower = direction.lower()
            if lower == "find":
                direction = self.FIND
            elif lower.startswith("min"):
                direction = self.MIN
            elif lower.startswith("max"):
                direction = self.MAX
            else:
                raise ValueError(
                    "Invalid search direction '{}'.".format(direction))

        if function is None:
            if direction != self.FIND:
                raise ValueError("Missing an objective function.")
        else:
            if direction == self.FIND:
                raise ValueError("May not specify an objective function for a "
                    "feasiblity problem.")

            if not isinstance(function, expressions.Expression):
                raise TypeError(
                    "Objective function must be a PICOS expression.")

            if len(function) != 1:
                raise TypeError("Objective function must be scalar.")

            function = function.refined

            if isinstance(function, expressions.ComplexAffineExpression) \
            and function.complex:
                raise TypeError("Objective function may not be complex.")

        self._direction = direction
        self._function = function

    def __str__(self):
        if self._function is None:
            return "find a variable assignment"
        else:
            dir_str = "minimize" if self._direction == self.MIN else "maximize"
            return "{} {}".format(dir_str, self._function.string)

    def __repr__(self):
        return glyphs.repr1("Objective: {}".format(self))

    def __iter__(self):
        yield self._direction
        yield self._function

    def __eq__(self, other):
        if not isinstance(other, Objective):
            return False

        if self._direction != other._direction:
            return False

        if self._direction == self.FIND:
            return True

        try:
            return self._function.equals(other._function)
        except AttributeError:
            # TODO: Allow all expressions to be equality-checked?
            return self._function is other._function

    @property
    def feasibility(self):
        """Whether the objective is "find an assignment"."""
        return self._function is None

    @property
    def pair(self):
        """Search direction and objective function as a pair."""
        return self._direction, self._objective

    @property
    def direction(self):
        """Search direction as a short string."""
        return self._direction

    @property
    def function(self):
        """Objective function."""
        return self._function

    @cached_property
    def normalized(self):
        """The objective but with feasiblity posed as "minimize 0".

        >>> from picos import Objective
        >>> obj = Objective(); obj
        <Objective: find a variable assignment>
        >>> obj.normalized
        <Objective: minimize 0>
        """
        if self._function is None:
            return Objective(self.MIN, expressions.AffineExpression.zero())
        else:
            return self

    @property
    def value(self):
        """Value of the objective function."""
        if self._function is None:
            return None
        else:
            return self._function.value

    def __index__(self):
        if self._function is None:
            raise TypeError("A feasiblity objective cannot be used as an index "
                "because there is no objective function to take the value of.")

        value = self.value

        if value is None:
            raise expressions.NotValued(
                "Cannot use unvalued objective function {} as an index."
                .format(self._function.string))

        try:
            fltValue = float(value)
            intValue = int(fltValue)

            if intValue != fltValue:
                raise ValueError
        except (TypeError, ValueError):
            raise RuntimeError("Cannot use the objective function {} as an "
                "index as its value of {} is not integral."
                .format(self._function.string, value))

        return intValue

    def _casting_helper(self, theType):
        assert theType in (int, float, complex)

        if self._function is None:
            raise TypeError("A feasiblity objective cannot be cast as {} "
                "because there is no objective function to take the value of."
                .format(theType.__name__))

        value = self.value

        if value is None:
            raise expressions.NotValued(
                "Cannot cast unvalued objective function {} as {}."
                .format(self._function.string, theType.__name__))

        return theType(value)

    def __int__(self):
        return self._casting_helper(int)

    def __long__(self):
        # Python 2 only.
        return self._casting_helper(int).__long__()

    def __float__(self):
        return self._casting_helper(float)

    def __complex__(self):
        return self._casting_helper(complex)

    def __round__(self, ndigits=None):
        return round(float(self), ndigits)


class Problem(object):  # object inheritence necessary with Python 2.
    """PICOS' representation of an optimization problem.

    :Example:

    >>> from picos import Problem, RealVariable
    >>> X = RealVariable("X", (2,2), lower = 0)
    >>> P = Problem()
    >>> P.set_objective("max", X.tr)
    >>> C1 = P.add_constraint(X.sum <= 10)
    >>> C2 = P.add_constraint(X[0,0] == 1)
    >>> print(P)
    Linear Program
      maximize tr(X)
      over
        2×2 real variable X (bounded below)
      subject to
        ∑(X) ≤ 10
        X[0,0] = 1
    >>> # PICOS will select a suitable solver if you don't specify one.
    >>> solution = P.solve(solver = "cvxopt")
    >>> solution.claimedStatus
    'optimal'
    >>> solution.searchTime #doctest: +SKIP
    0.002137422561645508
    >>> round(P, 1)
    10.0
    >>> print(X) #doctest: +SKIP
    [ 1.00e+00  4.89e-10]
    [ 4.89e-10  9.00e+00]
    >>> round(C1.dual, 1)
    1.0
    """

    # --------------------------------------------------------------------------
    # Initialization and reset methods.
    # --------------------------------------------------------------------------

    def __init__(self, copyOptions=None, useOptions=None, **extra_options):
        """Create an empty problem and optionally set initial solver options.

        :param copyOptions:
            An :class:`Options <picos.Options>` object to copy instead of using
            the default options.

        :param useOptions: An :class:`Options <picos.Options>` object to use
            (without making a copy) instead of using the default options.

        :param extra_options:
            A sequence of additional solver options to apply on top of the
            default options or those given by ``copyOptions`` or ``useOptions``.
        """
        if copyOptions and useOptions:
            raise ValueError(
                "Can only copy or use existing solver options, not both.")

        extra_options = map_legacy_options(**extra_options)

        if copyOptions:
            self._options = copyOptions.copy()
            self._options.update(**extra_options)
        elif useOptions:
            self._options = useOptions
            self._options.update(**extra_options)
        else:
            self._options = Options(**extra_options)

        #: The optimization objective.
        self._objective = Objective()

        #: Maps constraint IDs to constraints.
        self._constraints = OrderedDict()

        #: Contains lists of constraints added together, all in order.
        self._con_groups = []

        #: Maps variable names to variables.
        self._variables = OrderedDict()

        #: Maps variables to number of occurences in objective or constraints.
        self._var_count = {}

        #: Current solution strategy.
        self._strategy = None

        #: The last :class:`Solution` applied to the problem.
        self._last_solution = None  # Set by Solution.apply.

    def reset(self, resetOptions=False):
        """Reset the problem instance to its initial empty state.

        :param bool resetOptions:
            Whether also solver options should be reset to their default values.
        """
        # Reset options if requested.
        if resetOptions:
            self._options.reset()

        # Reset objective to "find an assignment".
        del self.objective

        # Reset constraint registry.
        self._constraints.clear()
        self._con_groups.clear()

        # Reset variable registry.
        self._variables.clear()
        self._var_count.clear()

        # Reset strategy and solution data.
        self._strategy = None
        self._last_solution = None

    # --------------------------------------------------------------------------
    # Properties.
    # --------------------------------------------------------------------------

    @property
    def variables(self):
        """Maps names to variables in use by the problem.

        :returns:
            A read-only view to an :class:`~collections.OrderedDict`. The order
            is deterministic and depends on the order of operations performed on
            the :class:`Problem` instance as well as on the variables' names.
        """
        return MappingProxyType(self._variables)

    @property
    def constraints(self):
        """Maps constraint IDs to constraints that are part of the problem.

        :returns:
            A read-only view to an :class:`~collections.OrderedDict`. The order
            is that in which constraints were added.
        """
        return MappingProxyType(self._constraints)

    @constraints.deleter
    def constraints(self):
        # Clear constraint registry.
        self._constraints.clear()
        self._con_groups.clear()

        # Clear variable registry except for the objective function variables.
        self._variables.clear()
        self._var_count.clear()
        self._register_variables(self.no.function.variables)

    @property
    def objective(self):
        """Optimization objective as an :class:`~picos.Objective` instance."""
        return self._objective

    @objective.setter
    def objective(self, value):
        self._unregister_variables(self.no.function.variables)

        try:
            if isinstance(value, Objective):
                self._objective = value
            else:
                direction, function = value
                self._objective = Objective(direction, function)
        finally:
            self._register_variables(self.no.function.variables)

    @objective.deleter
    def objective(self):
        self._unregister_variables(self.no.function.variables)

        self._objective = Objective()

    @property
    def no(self):
        """Normalized objective as an :class:`~picos.Objective` instance.

        Either a minimization or a maximization objective, with feasibility
        posed as "minimize 0".

        The same as the :attr:`~picos.Objective.normalized` attribute of the
        :attr:`objective`.
        """
        return self._objective.normalized

    @property
    def value(self):
        """Objective function value.

        If all variables that appear in the objective function are valued, in
        particular after a successful solution search, this is the numeric value
        of the objective function. If the objective function is not fully valued
        or if the problem is a feasiblity problem without an objective function,
        this is :obj:`None`.

        .. note::
            The Python special functions :class:`int`, :class:`float`,
            :class:`complex` and :func:`round` as well as the special method
            ``__index__`` make use of this value when applied to a
            :class:`Problem`.
        """
        return self._objective.value

    @property
    def options(self):
        """Solution search parameters as an :class:`~picos.Options` object."""
        return self._options

    @options.setter
    def options(self, value):
        if not isinstance(value, Options):
            raise TypeError("Cannot assign an object of type {} as a problem's "
                " options.".format(type(value).__name__))

        self._options = value

    @options.deleter
    def options(self, value):
        self._options.reset()

    @property
    def strategy(self):
        """Solution strategy as a :class:`~picos.modeling.Strategy` object.

        A strategy is available once you order the problem to be solved and it
        will be reused for successive solution attempts (of a modified problem)
        while it remains valid with respect to the problem's :attr:`footprint`.

        When a strategy is reused, modifications to the variables, constraints,
        and objective of a problem are passed step by step through the
        strategy's reformulation pipeline while existing reformulation work is
        not repeated. If the solver also supports these kinds of updates, then
        modifying and re-solving a problem can be much faster than solving the
        problem from scratch.

        :Example:

        >>> from picos import Problem, RealVariable
        >>> x = RealVariable("x")
        >>> P = Problem()
        >>> P.set_objective("min", x**2)
        >>> print(P.strategy)
        None
        >>> sol = P.solve(solver = "cvxopt")  # Creates a solution strategy.
        >>> print(P.strategy)
        1. ExtraOptions
        2. EpigraphReformulation
        3. ConvexQuadraticToConicReformulation
        4. CVXOPTSolver
        >>> P.add_constraint((x-2)**2 <= 1)  # Also convex quadratic.
        <Convex Quadratic Constraint: (x - 2)² ≤ 1>
        >>> P.strategy.valid(solver = "cvxopt")
        True
        >>> P.strategy.valid(solver = "glpk")
        False
        >>> sol = P.solve(solver = "cvxopt")  # Reuses the strategy.

        It's also possible to create a startegy from scratch:

        >>> from picos.modeling import Strategy
        >>> from picos.reforms import (EpigraphReformulation,
        ...     ConvexQuadraticToConicReformulation)
        >>> from picos.solvers import CVXOPTSolver
        >>> # Mimic what solve() does when no strategy exists:
        >>> P.strategy = Strategy(P, CVXOPTSolver, EpigraphReformulation,
        ...     ConvexQuadraticToConicReformulation)
        """
        return self._strategy

    @strategy.setter
    def strategy(self, value):
        from .strategy import Strategy

        if not isinstance(value, Strategy):
            raise TypeError(
                "Cannot assign an object of type {} as a solution strategy."
                .format(type(value).__name__))

        if value.problem is not self:
            raise ValueError("The solution strategy was constructed for a "
                "different problem.")

        self._strategy = value

    @strategy.deleter
    def strategy(self):
        self._strategy = None

    @property
    def last_solution(self):
        """The last :class:`~picos.Solution` applied to the problem."""
        return self._last_solution

    @property
    def status(self):
        """The solution status string as claimed by :attr:`last_solution`."""
        if not self._last_solution:
            return "unsolved"
        else:
            return self._last_solution.claimedStatus

    @property
    def footprint(self):
        """Problem footprint as a :class:`~picos.modeling.Footprint` object."""
        return Footprint.from_problem(self)

    @property
    def continuous(self):
        """Whether all variables are of continuous types."""
        return all(
            isinstance(variable, expressions.CONTINUOUS_VARTYPES)
            for variable in self.variables.values())

    @property
    def pure_integer(self):
        """Whether all variables are of integral types."""
        return not any(
            isinstance(variable, expressions.CONTINUOUS_VARTYPES)
            for variable in self.variables.values())

    @property
    def type(self):
        """The problem type as a string, such as "Linear Program"."""
        C = set(type(c) for c in self._constraints.values())
        objective = self._objective.function
        base = "Optimization Problem"

        linear = [
            constraints.AffineConstraint,
            constraints.ComplexAffineConstraint,
            constraints.AbsoluteValueConstraint,
            constraints.SimplexConstraint,
            constraints.FlowConstraint]
        sdp = [
            constraints.LMIConstraint,
            constraints.ComplexLMIConstraint]
        quadratic = [
            constraints.ConvexQuadraticConstraint,
            constraints.ConicQuadraticConstraint,
            constraints.NonconvexQuadraticConstraint]
        quadconic = [
            constraints.SOCConstraint,
            constraints.RSOCConstraint]
        exponential = [
            constraints.ExpConeConstraint,
            constraints.SumExponentialsConstraint,
            constraints.LogSumExpConstraint,
            constraints.LogConstraint,
            constraints.KullbackLeiblerConstraint]
        complex = [
            constraints.ComplexAffineConstraint,
            constraints.ComplexLMIConstraint]

        if objective is None:
            if not C:
                base = "Empty Problem"
            elif C.issubset(set(linear)):
                base = "Linear Feasibility Problem"
            else:
                base = "Feasibility Problem"
        elif isinstance(objective, expressions.AffineExpression):
            if not C:
                if objective.constant:
                    base = "Constant Problem"
                else:
                    base = "Linear Program"  # Could have variable bounds.
            elif C.issubset(set(linear)):
                base = "Linear Program"
            elif C.issubset(set(linear + quadconic)):
                base = "Second Order Cone Program"
            elif C.issubset(set(linear + sdp)):
                base = "Semidefinite Program"
            elif C.issubset(set(linear + [constraints.LogSumExpConstraint])):
                base = "Geometric Program"
            elif C.issubset(set(linear + exponential)):
                base = "Exponential Program"
            elif C.issubset(set(linear + quadratic)):
                base = "Quadratically Constrained Program"
        elif isinstance(objective, expressions.QuadraticExpression):
            if C.issubset(set(linear)):
                base = "Quadratic Program"
            elif C.issubset(set(linear + quadratic)):
                base = "Quadratically Constrained Quadratic Program"
        elif isinstance(objective, expressions.LogSumExp):
            if C.issubset(set(linear + [constraints.LogSumExpConstraint])):
                base = "Geometric Program"

        if self.continuous:
            integrality = ""
        elif self.pure_integer:
            integrality = "Integer "
        else:
            integrality = "Mixed-Integer "

        if any(c in complex for c in C):
            complexity = "Complex "
        else:
            complexity = ""

        return "{}{}{}".format(complexity, integrality, base)

    @property
    def dual(self):
        """The Lagrangian dual problem of the standardized problem.

        More precisely, this property invokes the following:

        1. The primal problem is posed as an equivalent conic standard form
           minimization problem, with variable bounds expressed as additional
           constraints.
        2. The Lagrangian dual problem of the reposed primal is computed.
        3. The optimization direction and objective function sign of the dual
           are adjusted such that, given strong duality and primal feasibility,
           the optimal values of both problems are equal. In particular, if the
           primal problem is a minimization or a maximization problem, the dual
           problem returned will be the respective other.

        This property is intended for educational purposes. If you want to solve
        the primal problem via its dual, use the :ref:`dualize <option_dualize>`
        option instead.
        """
        return self.prepared(solver="cvxopt", dualize=True)

    # --------------------------------------------------------------------------
    # Python special methods, except __init__.
    # --------------------------------------------------------------------------

    @property
    def _var_groups(self):
        """Support :meth:`_var_group_string`."""
        vars_by_type = {}
        for var in self.variables.values():
            vtype = type(var).__name__
            shape = var.shape
            bound = tuple(bool(bound) for bound in var.bound_dicts)
            index = (vtype, shape, bound)

            vars_by_type.setdefault(index, set())
            vars_by_type[index].add(var)

        groups = []
        for index in sorted(vars_by_type.keys()):
            groups.append(natsorted(vars_by_type[index], key=lambda v: v.name))

        return groups

    @lru_cache()
    def _var_group_string(self, group):
        """Support :meth:`__str__`."""
        if len(group) == 0:
            return "[no variables]"

        if len(group) == 1:
            return group[0].long_string

        try:
            template, data = parameterized_string(
                [c.long_string for c in group])
        except ValueError:
            return group[0].long_string \
                + ", " + ", ".join([v.name for v in group[1:]])
        else:
            return glyphs.forall(template, data)

    @lru_cache()
    def _con_group_string(self, group):
        """Support :meth:`__str__`."""
        if len(group) == 0:
            return "[no constraints]"

        if len(group) == 1:
            return str(group[0])

        try:
            template, data = parameterized_string([str(c) for c in group])
        except ValueError:
            return "[{} constraints (1st: {})]".format(len(group), group[0])
        else:
            return glyphs.forall(template, data)

    def __repr__(self):
        return glyphs.repr1(self.type)

    def __str__(self):
        # Print problem type.
        string = "{}\n".format(self.type)

        # Print objective.
        objDir, obj = self._objective
        if objDir == "max":
            string += "  maximize {}\n".format(obj.string)
        elif objDir == "min":
            string += "  minimize {}\n".format(obj.string)
        elif objDir == "find":
            if self._constraints:
                string += "  find an assignment\n"
        else:
            assert False, "Unexpected objective direction '{}'.".format(objDir)

        wrapper = TextWrapper(
            initial_indent=" "*4,
            subsequent_indent=" "*6,
            break_long_words=False,
            break_on_hyphens=False)

        # Print variables.
        if self.variables:
            string += "  {}\n".format("for" if objDir == "find" else "over")
            for group in self._var_groups:
                string += wrapper.fill(self._var_group_string(tuple(group)))
                string += "\n"

        # Print constraints.
        if self._constraints:
            string += "  subject to\n"
            for index, group in enumerate(self._con_groups):
                string += wrapper.fill(self._con_group_string(tuple(group)))
                string += "\n"

        return string.rstrip("\n")

    def __index__(self):
        return self._objective.__index__()

    def __int__(self):
        return self._objective.__int__()

    def __long__(self):
        # Python 2 only.
        return self._objective.__long__()

    def __float__(self):
        return self._objective.__float__()

    def __complex__(self):
        return self._objective.__complex__()

    def __round__(self, ndigits=None):
        return self._objective.__round__(ndigits)

    # --------------------------------------------------------------------------
    # Bookkeeping methods.
    # --------------------------------------------------------------------------

    def _register_variables(self, vars):
        """Register the variables of an objective function or constraint."""
        # Register every variable at most once per call.
        if not isinstance(vars, (set, frozenset)):
            raise TypeError("Variable registry can (un)register a variable "
                "only once per call, so the argument must be a set type.")

        # Retrieve old and new variables as mapping from name to object.
        oldVars = self._variables
        newVars = OrderedDict(
            (var.name, var) for var in sorted(vars, key=(lambda v: v.name)))

        # Check for variable name clashes within the new set.
        if len(newVars) != len(vars):
            raise ValueError(
                "The object you are trying to add to a problem contains "
                "multiple variables of the same name. This is not allowed.")

        # Check for variable name clashes with existing variables.
        for name in set(oldVars).intersection(set(newVars)):
            if oldVars[name] is not newVars[name]:
                raise ValueError("Cannot register the variable {} with the "
                    "problem because it already tracks another variable with "
                    "the same name.".format(name))

        # Keep track of new variables.
        self._variables.update(newVars)

        # Count up the variable references.
        for var in vars:
            self._var_count.setdefault(var, 0)
            self._var_count[var] += 1

    def _unregister_variables(self, vars):
        """Unregister the variables of an objective function or constraint."""
        # Unregister every variable at most once per call.
        if not isinstance(vars, (set, frozenset)):
            raise TypeError("Variable registry can (un)register a variable "
                "only once per call, so the argument must be a set type.")

        for var in vars:
            name = var.name

            # Make sure the variable is properly registered.
            assert name in self._variables and var in self._var_count, \
                "Tried to unregister a variable that is not registered."
            assert self._var_count[var] >= 1, \
                "Found a nonpostive variable count."

            # Count down the variable references.
            self._var_count[var] -= 1

            # Remove a variable with a reference count of zero.
            if not self._var_count[var]:
                self._var_count.pop(var)
                self._variables.pop(name)

    # --------------------------------------------------------------------------
    # Methods to manipulate the objective function and its direction.
    # --------------------------------------------------------------------------

    def set_objective(self, direction=None, expression=None):
        """Set the optimization direction and objective function of the problem.

        :param str direction:
            Case insensitive search direction string. One of

            - ``"min"`` or ``"minimize"``,
            - ``"max"`` or ``"maximize"``,
            - ``"find"`` or :obj:`None` (for a feasibility problem).

        :param ~picos.expressions.Expression expression:
            The objective function. Must be :obj:`None` for a feasibility
            problem.
        """
        self.objective = direction, expression

    # --------------------------------------------------------------------------
    # Methods to add, retrieve and remove constraints.
    # --------------------------------------------------------------------------

    def _lookup_constraint(self, idOrIndOrCon):
        """Look for a constraint with the given identifier.

        Given a constraint object or ID or offset or a constraint group index or
        index pair, returns a matching (list of) constraint ID(s) that is (are)
        part of the problem.
        """
        if isinstance(idOrIndOrCon, int):
            if idOrIndOrCon in self._constraints:
                # A valid ID.
                return idOrIndOrCon
            elif idOrIndOrCon < len(self._constraints):
                # An offset.
                return list(self._constraints.keys())[idOrIndOrCon]
            else:
                raise LookupError(
                    "The problem has no constraint with ID or offset {}."
                    .format(idOrIndOrCon))
        elif isinstance(idOrIndOrCon, constraints.Constraint):
            # A constraint object.
            id = idOrIndOrCon.id
            if id in self._constraints:
                return id
            else:
                raise KeyError("The constraint '{}' is not part of the problem."
                    .format(idOrIndOrCon))
        elif isinstance(idOrIndOrCon, tuple) or isinstance(idOrIndOrCon, list):
            if len(idOrIndOrCon) == 1:
                groupIndex = idOrIndOrCon[0]
                if groupIndex < len(self._con_groups):
                    return [c.id for c in self._con_groups[groupIndex]]
                else:
                    raise IndexError("Constraint group index out of range.")
            elif len(idOrIndOrCon) == 2:
                groupIndex, groupOffset = idOrIndOrCon
                if groupIndex < len(self._con_groups):
                    group = self._con_groups[groupIndex]
                    if groupOffset < len(group):
                        return group[groupOffset].id
                    else:
                        raise IndexError(
                            "Constraint group offset out of range.")
                else:
                    raise IndexError("Constraint group index out of range.")
            else:
                raise TypeError("If looking up constraints by group, the index "
                    "must be a tuple or list of length at most two.")
        else:
            raise TypeError("Argument of type '{}' not supported when looking "
                "up constraints".format(type(idOrIndOrCon)))

    def get_constraint(self, idOrIndOrCon):
        """Return a (list of) constraint(s) of the problem.

        :param idOrIndOrCon: One of the following:

            * A constraint object. It will be returned when the constraint is
              part of the problem, otherwise a KeyError is raised.
            * The integer ID of the constraint.
            * The integer offset of the constraint in the list of all
              constraints that are part of the problem, in the order that they
              were added.
            * A list or tuple of length 1. Its only element is the index of a
              constraint group (of constraints that were added together), where
              groups are indexed in the order that they were added to the
              problem. The whole group is returned as a list of constraints.
              That list has the constraints in the order that they were added.
            * A list or tuple of length 2. The first element is a constraint
              group offset as above, the second an offset within that list.

        :type idOrIndOrCon: picos.constraints.Constraint or int or tuple or list

        :returns: A :class:`constraint <picos.constraints.Constraint>` or a list
            thereof.

        :Example:

        >>> import picos as pic
        >>> import cvxopt as cvx
        >>> from pprint import pprint
        >>> prob=pic.Problem()
        >>> x=[prob.add_variable('x[{0}]'.format(i),2) for i in range(5)]
        >>> y=prob.add_variable('y',5)
        >>> Cx=prob.add_list_of_constraints([(1|x[i]) < y[i] for i in range(5)])
        >>> Cy=prob.add_constraint(y>0)
        >>> print(prob)
        Linear Feasibility Problem
          find an assignment
          for
            2×1 real variable x[i] ∀ i ∈ [0…4]
            5×1 real variable y
          subject to
            ∑(x[i]) ≤ y[i] ∀ i ∈ [0…4]
            y ≥ 0
        >>> # Retrieve the second constraint, indexed from zero:
        >>> prob.get_constraint(1)
        <1×1 Affine Constraint: ∑(x[1]) ≤ y[1]>
        >>> # Retrieve the fourth consraint from the first group:
        >>> prob.get_constraint((0,3))
        <1×1 Affine Constraint: ∑(x[3]) ≤ y[3]>
        >>> # Retrieve the whole first group of constraints:
        >>> pprint(prob.get_constraint((0,)))
        [<1×1 Affine Constraint: ∑(x[0]) ≤ y[0]>,
         <1×1 Affine Constraint: ∑(x[1]) ≤ y[1]>,
         <1×1 Affine Constraint: ∑(x[2]) ≤ y[2]>,
         <1×1 Affine Constraint: ∑(x[3]) ≤ y[3]>,
         <1×1 Affine Constraint: ∑(x[4]) ≤ y[4]>]
        >>> # Retrieve the second "group", containing just one constraint:
        >>> prob.get_constraint((1,))
        [<5×1 Affine Constraint: y ≥ 0>]
        """
        idOrIds = self._lookup_constraint(idOrIndOrCon)

        if isinstance(idOrIds, list):
            return [self._constraints[id] for id in idOrIds]
        else:
            return self._constraints[idOrIds]

    def add_constraint(self, constraint, key=None):
        """Add a constraint to the problem.

        :param constraint: The constraint to be added.
        :type constraint: :class:`Constraint <picos.constraints.Constraint>`
        :param str key: Optional name of the constraint.
        :returns: The constraint that was added to the problem.
        """
        # Handle deprecated 'key' parameter.
        if key is not None:
            throw_deprecation_warning(
                "Naming constraints is currently not supported.")

        # Register the constraint.
        self._constraints[constraint.id] = constraint
        self._con_groups.append([constraint])

        # Register the constraint's variables.
        self._register_variables(constraint.variables)

        return constraint

    def add_list_of_constraints(self, lst, it=None, indices=None, key=None):
        """Add a list of constraints to the problem.

        Adds a list of constraints to the problem, enabling the use of
        Python list comprehensions (see the example below).

        :param list(picos.constraints.Constraint) lst: Constraints to add.
        :param it: DEPRECATED
        :param indices: DEPRECATED
        :param key: DEPRECATED

        :returns: A list of all constraints that were added.

        :Example:

        >>> import picos as pic
        >>> import cvxopt as cvx
        >>> from pprint import pprint
        >>> prob=pic.Problem()
        >>> x=[prob.add_variable('x[{0}]'.format(i),2) for i in range(5)]
        >>> pprint(x)
        [<2×1 Real Variable: x[0]>,
         <2×1 Real Variable: x[1]>,
         <2×1 Real Variable: x[2]>,
         <2×1 Real Variable: x[3]>,
         <2×1 Real Variable: x[4]>]
        >>> y=prob.add_variable('y',5)
        >>> IJ=[(1,2),(2,0),(4,2)]
        >>> w={}
        >>> for ij in IJ:
        ...         w[ij]=prob.add_variable('w[{},{}]'.format(*ij),3)
        ...
        >>> u=pic.new_param('u',cvx.matrix([2,5]))
        >>> C1=prob.add_list_of_constraints([u.T*x[i] < y[i] for i in range(5)])
        >>> C2=prob.add_list_of_constraints([abs(w[i,j])<y[j] for (i,j) in IJ])
        >>> C3=prob.add_list_of_constraints([y[t] > y[t+1] for t in range(4)])
        >>> print(prob)
        Feasibility Problem
          find an assignment
          for
            2×1 real variable x[i] ∀ i ∈ [0…4]
            3×1 real variable w[i,j] ∀ (i,j) ∈ zip([1,2,4],[2,0,2])
            5×1 real variable y
          subject to
            uᵀ·x[i] ≤ y[i] ∀ i ∈ [0…4]
            ‖w[i,j]‖ ≤ y[j] ∀ (i,j) ∈ zip([1,2,4],[2,0,2])
            y[i] ≥ y[i+1] ∀ i ∈ [0…3]
        """
        if it is not None or indices is not None or key is not None:
            # Deprecated as of 2.0.
            throw_deprecation_warning("Arguments 'it', 'indices' and 'key' to "
                "add_list_of_constraints are deprecated and ignored.")

        added = []
        for constraint in lst:
            added.append(self.add_constraint(constraint))
            self._con_groups.pop()

        if added:
            self._con_groups.append(added)

        return added

    # TODO: Add Problem.require replacing add_constraint and
    #       add_list_of_constraints?

    def _con_group_index(self, conOrConID):
        """Support :meth:`remove_constraint`."""
        if isinstance(conOrConID, int):
            constraint = self._constraints[conOrConID]
        else:
            constraint = conOrConID

        for i, group in enumerate(self._con_groups):
            for j, candidate in enumerate(group):
                if candidate is constraint:
                    return i, j

        if constraint in self._constraints.values():
            raise RuntimeError("The problem's constraint and constraint group "
                "registries are out of sync.")
        else:
            raise KeyError("The constraint is not part of the problem.")

    def remove_constraint(self, idOrIndOrCon):
        """Delete a constraint from the problem.

        :param idOrIndOrCon: See :meth:`get_constraint`.

        :Example:

        >>> import picos
        >>> from pprint import pprint
        >>> P = picos.Problem()
        >>> x = [P.add_variable('x[{0}]'.format(i), 2) for i in range(4)]
        >>> y = P.add_variable('y', 4)
        >>> Cxy = P.add_list_of_constraints(
        ...     [(1 | x[i]) <= y[i] for i in range(4)])
        >>> Cy = P.add_constraint(y >= 0)
        >>> Cx0to2 = P.add_list_of_constraints([x[i] <= 2 for i in range(3)])
        >>> Cx3 = P.add_constraint(x[3] <= 1)
        >>> pprint(list(P.constraints.values()))#doctest: +NORMALIZE_WHITESPACE
        [<1×1 Affine Constraint: ∑(x[0]) ≤ y[0]>,
         <1×1 Affine Constraint: ∑(x[1]) ≤ y[1]>,
         <1×1 Affine Constraint: ∑(x[2]) ≤ y[2]>,
         <1×1 Affine Constraint: ∑(x[3]) ≤ y[3]>,
         <4×1 Affine Constraint: y ≥ 0>,
         <2×1 Affine Constraint: x[0] ≤ [2]>,
         <2×1 Affine Constraint: x[1] ≤ [2]>,
         <2×1 Affine Constraint: x[2] ≤ [2]>,
         <2×1 Affine Constraint: x[3] ≤ [1]>]
        >>> # Delete the 2nd constraint (counted from 0):
        >>> P.remove_constraint(1)
        >>> pprint(list(P.constraints.values()))#doctest: +NORMALIZE_WHITESPACE
        [<1×1 Affine Constraint: ∑(x[0]) ≤ y[0]>,
         <1×1 Affine Constraint: ∑(x[2]) ≤ y[2]>,
         <1×1 Affine Constraint: ∑(x[3]) ≤ y[3]>,
         <4×1 Affine Constraint: y ≥ 0>,
         <2×1 Affine Constraint: x[0] ≤ [2]>,
         <2×1 Affine Constraint: x[1] ≤ [2]>,
         <2×1 Affine Constraint: x[2] ≤ [2]>,
         <2×1 Affine Constraint: x[3] ≤ [1]>]
        >>> # Delete the 2nd group of constraints, i.e. the constraint y > 0:
        >>> P.remove_constraint((1,))
        >>> pprint(list(P.constraints.values()))#doctest: +NORMALIZE_WHITESPACE
        [<1×1 Affine Constraint: ∑(x[0]) ≤ y[0]>,
         <1×1 Affine Constraint: ∑(x[2]) ≤ y[2]>,
         <1×1 Affine Constraint: ∑(x[3]) ≤ y[3]>,
         <2×1 Affine Constraint: x[0] ≤ [2]>,
         <2×1 Affine Constraint: x[1] ≤ [2]>,
         <2×1 Affine Constraint: x[2] ≤ [2]>,
         <2×1 Affine Constraint: x[3] ≤ [1]>]
        >>> # Delete the 3rd remaining group of constraints, i.e. x[3] < [1]:
        >>> P.remove_constraint((2,))
        >>> pprint(list(P.constraints.values()))#doctest: +NORMALIZE_WHITESPACE
        [<1×1 Affine Constraint: ∑(x[0]) ≤ y[0]>,
         <1×1 Affine Constraint: ∑(x[2]) ≤ y[2]>,
         <1×1 Affine Constraint: ∑(x[3]) ≤ y[3]>,
         <2×1 Affine Constraint: x[0] ≤ [2]>,
         <2×1 Affine Constraint: x[1] ≤ [2]>,
         <2×1 Affine Constraint: x[2] ≤ [2]>]
        >>> # Delete 2nd constraint of the 2nd remaining group, i.e. x[1] < |2|:
        >>> P.remove_constraint((1,1))
        >>> pprint(list(P.constraints.values()))#doctest: +NORMALIZE_WHITESPACE
        [<1×1 Affine Constraint: ∑(x[0]) ≤ y[0]>,
         <1×1 Affine Constraint: ∑(x[2]) ≤ y[2]>,
         <1×1 Affine Constraint: ∑(x[3]) ≤ y[3]>,
         <2×1 Affine Constraint: x[0] ≤ [2]>,
         <2×1 Affine Constraint: x[2] ≤ [2]>]
        """
        idOrIds = self._lookup_constraint(idOrIndOrCon)

        removedCons = []

        if isinstance(idOrIds, list):
            assert idOrIds, "There is an empty constraint group."
            groupIndex, _ = self._con_group_index(idOrIds[0])
            self._con_groups.pop(groupIndex)
            for id in idOrIds:
                removedCons.append(self._constraints.pop(id))
        else:
            constraint = self._constraints.pop(idOrIds)
            removedCons.append(constraint)
            groupIndex, groupOffset = self._con_group_index(constraint)
            group = self._con_groups[groupIndex]
            group.pop(groupOffset)
            if not group:
                self._con_groups.pop(groupIndex)

        # Unregister the variables added by the removed constraints.
        for con in removedCons:
            self._unregister_variables(con.variables)

    def remove_all_constraints(self):
        """Remove all constraints from the problem.

        .. note::

            This method does not remove bounds set directly on variables.
        """
        del self.constraints

    # --------------------------------------------------------------------------
    # Borderline legacy methods to deal with variables.
    # TODO: Document removal of Problem.remove_all_variable_bounds.
    # --------------------------------------------------------------------------

    _PARAMETERIZED_VARIABLE_REGEX = re.compile(r"^([^[]+)\[([^\]]+)\]$")

    def get_variable(self, name):
        """Retrieve variables referenced by the problem.

        Retrieves either a single variable with the given name or a group of
        variables all named ``name[param]`` with different values for ``param``.
        If the values for ``param`` are the integers from zero to the size of
        the group minus one, then the group is returned as a :obj:`list` ordered
        by ``param``, otherwise it is returned as a :obj:`dict` with the values
        of ``param`` as keys.

        .. note::

            Since PICOS 2.0, variables are independent of problems and only
            appear in a problem for as long as they are referenced by the
            problem's objective function or constraints.

        :param str name:
            The name of a variable, or the base name of a group of variables.

        :returns:
            A :class:`variable <picos.expressions.BaseVariable>` or a
            :class:`list` or :class:`dict` thereof.

        :Example:

        >>> from picos import Problem, RealVariable
        >>> from pprint import pprint
        >>> # Create a number of variables with structured names.
        >>> vars = [RealVariable("x")]
        >>> for i in range(4):
        ...     vars.append(RealVariable("y[{}]".format(i)))
        >>> for key in ["alice", "bob", "carol"]:
        ...     vars.append(RealVariable("z[{}]".format(key)))
        >>> # Make the variables appear in a problem.
        >>> P = Problem()
        >>> P.set_objective("min", sum([var for var in vars]))
        >>> print(P)
        Linear Program
          minimize x + y[0] + y[1] + y[2] + y[3] + z[alice] + z[bob] + z[carol]
          over
            1×1 real variable x, y[0], y[1], y[2], y[3], z[alice], z[bob],
              z[carol]
        >>> # Retrieve the variables from the problem.
        >>> P.get_variable("x")
        <1×1 Real Variable: x>
        >>> pprint(P.get_variable("y"))
        [<1×1 Real Variable: y[0]>,
         <1×1 Real Variable: y[1]>,
         <1×1 Real Variable: y[2]>,
         <1×1 Real Variable: y[3]>]
        >>> pprint(P.get_variable("z"))
        {'alice': <1×1 Real Variable: z[alice]>,
         'bob': <1×1 Real Variable: z[bob]>,
         'carol': <1×1 Real Variable: z[carol]>}
        >>> P.get_variable("z")["alice"] is P.get_variable("z[alice]")
        True
        """
        if name in self._variables:
            return self._variables[name]
        else:
            # Check if the name is really just a basename.
            params = []
            for otherName in sorted(self._variables.keys()):
                match = self._PARAMETERIZED_VARIABLE_REGEX.match(otherName)
                if not match:
                    continue
                base, param = match.groups()
                if name == base:
                    params.append(param)

            if params:
                # Return a list if the parameters are a range.
                try:
                    intParams = sorted([int(p) for p in params])
                except ValueError:
                    pass
                else:
                    if intParams == list(range(len(intParams))):
                        return [self._variables["{}[{}]".format(name, param)]
                            for param in intParams]

                # Otherwise return a dict.
                return {param: self._variables["{}[{}]".format(name, param)]
                    for param in params}
            else:
                raise KeyError("The problem references no variable or group of "
                    "variables named '{}'.".format(name))

    def get_valued_variable(self, name):
        """Retrieve values of variables referenced by the problem.

        This method works the same :meth:`get_variable` but it returns the
        variable's :attr:`values <picos.expressions.Expression.value>` instead
        of the variable objects.

        :raises ~picos.expressions.NotValued:
            If any of the selected variables is not valued.
        """
        exp = self.get_variable(name)
        if isinstance(exp, list):
            for i in range(len(exp)):
                exp[i] = exp[i].value
        elif isinstance(exp, dict):
            for i in exp:
                exp[i] = exp[i].value
        else:
            exp = exp.value
        return exp

    # --------------------------------------------------------------------------
    # Methods to create copies of the problem.
    # --------------------------------------------------------------------------

    def copy(self):
        """Create a deep copy of the problem, using new variables."""
        theCopy = Problem(copyOptions=self._options)

        # Duplicate the variables.
        newVars = {name: var.copy() for name, var in self.variables.items()}

        # Make copies of constraints on top of the new variables.
        for group in self._con_groups:
            theCopy.add_list_of_constraints(
                constraint.replace_variables(newVars) for constraint in group)

        # Make a copy of the objective on top of the new variables.
        direction, function = self._objective
        if function is not None:
            theCopy.objective = direction, function.replace_variables(newVars)

        return theCopy

    def continuous_relaxation(self, copy_continuous_variables=True):
        """Return a continuous relaxation of the problem.

        This is done by replacing integer variables with continuous ones.

        :param bool copy_continuous_variables:
            Whether variables that are already continuous should still be
            copied. If this is :obj:`False`, then the relxation shares these
            variables with the original problem.
        """
        theCopy = Problem(copyOptions=self._options)

        # Duplicate the variables.
        newVars = {}
        for name, var in self.variables.items():
            if isinstance(var, expressions.CONTINUOUS_VARTYPES):
                if copy_continuous_variables:
                    newVars[name] = var.copy()
                else:
                    newVars[name] = var
            elif isinstance(var, expressions.IntegerVariable):
                newVars[name] = expressions.RealVariable(
                    name, var.shape, var._lower, var._upper)
            elif isinstance(var, expressions.BinaryVariable):
                newVars[name] = expressions.RealVariable(name, var.shape, 0, 1)

        # Make copies of constraints on top of the new variables.
        for group in self._con_groups:
            theCopy.add_list_of_constraints(
                constraint.replace_variables(newVars) for constraint in group)

        # Make a copy of the objective on top of the new variables.
        direction, function = self._objective
        if function is not None:
            theCopy.objective = direction, function.replace_variables(newVars)

        return theCopy

    def clone(self, copyOptions=True):
        """Create a semi-deep copy of the problem.

        The copy is constrained by the same constraint objects and thereby
        references the existing variables that appear in these constraints.

        The clone can be modified to describe a new problem, but whenever a
        solution is applied to its variables, some of these variables can be the
        same that still appear in the original problem. If this is not a problem
        to you, then cloning can be much faster than copying.

        :param bool copyOptions: Whether to make an independent copy of the
            problem's options. Disabling this will apply any option changes to
            the original problem as well but yields a (very small) reduction
            in cloning time.
        """
        # Start with a shallow copy of self.
        # TODO: Consider adding Problem.__new__ to speed this up further.
        theClone = pycopy.copy(self)

        # Make the constraint registry independent.
        theClone._constraints = self._constraints.copy()
        theClone._con_groups = []
        for group in self._con_groups:
            theClone._con_groups.append(pycopy.copy(group))

        # Make the variable registry independent.
        theClone._variables = self._variables.copy()
        theClone._var_count = self._var_count.copy()

        # Reset the clone's solution strategy and last solution.
        theClone._strategy = None

        # Make the solver options independent, if requested.
        if copyOptions:
            theClone._options = self._options.copy()

        # NOTE: No need to change the following attributes:
        #       - objective: Is immutable as a tuple.
        #       - _last_solution: Remains as valid as it is.

        return theClone

    # --------------------------------------------------------------------------
    # Methods to solve or export the problem.
    # --------------------------------------------------------------------------

    def prepared(self, steps=None, **extra_options):
        """Perform a dry-run returning the reformulated (prepared) problem.

        This behaves like :meth:`solve` in that it takes a number of additional
        temporary options, finds a solution strategy matching the problem and
        options, and performs the strategy's reformulations in turn to obtain
        modified problems. However, it stops after the given number of steps and
        never hands the reformulated problem to a solver. Instead of a solution,
        :meth:`prepared` then returns the last reformulated problem.

        Unless this method returns the problem itself, the special attributes
        ``prepared_strategy`` and ``prepared_steps`` are added to the returned
        problem. They then contain the (partially) executed solution strategy
        and the number of performed reformulations, respectively.

        :param int steps:
            Number of reformulations to perform. :obj:`None` means as many as
            there are. If this parameter is :math:`0`, then the problem itself
            is returned. If it is :math:`1`, then only the implicit first
            reformulation :meth:`~picos.reforms.ExtraOptions` is executed, which
            may also output the problem itself, depending on ``extra_options``.

        :param extra_options:
            Additional solver options to use with this dry-run only.

        :returns:
            The reformulated problem.

        :raises ~picos.modeling.strategy.NoStrategyFound:
            If no solution strategy was found.

        :raises ValueError:
            If there are not as many reformulation steps as requested.

        :Example:

        >>> from picos import Problem, RealVariable
        >>> x = RealVariable("x")
        >>> P = Problem()
        >>> P.set_objective("min", x**2)
        >>> Q = P.prepared(solver = "cvxopt")
        >>> print(Q.prepared_strategy)  # Show prepared reformulation steps.
        1. ExtraOptions
        2. EpigraphReformulation
        3. ConvexQuadraticToConicReformulation
        4. CVXOPTSolver
        >>> Q.prepared_steps  # Check how many steps have been performed.
        3
        >>> print(P)
        Quadratic Program
          minimize x²
          over
            1×1 real variable x
        >>> print(Q)# doctest: +ELLIPSIS
        Second Order Cone Program
          minimize __..._t
          over
            1×1 real variable __..._t, x
          subject to
            ‖quadroot(x²)‖² ≤ -(aff(x²) - __..._t) ∧ -(aff(x²) -
              __..._t) ≥ 0
        """
        from .strategy import Strategy

        # Produce a strategy for the clone.
        strategy = Strategy.from_problem(self, **extra_options)
        numReforms = len(strategy.reforms)

        if steps is None:
            steps = numReforms

        if steps == 0:
            return self
        elif steps > numReforms:
            raise ValueError("The pipeline {} has only {} reformulation steps "
                "to choose from.".format(strategy, numReforms, steps))

        # Replace the successor of the last reformulation with a dummy solver.
        lastReform = strategy.reforms[steps - 1]
        oldSuccessor = lastReform.successor
        lastReform.successor = type("DummySolver", (), {
            "execute": lambda self: Solution(
                {}, solver="dummy", vectorizedPrimals=True)})()

        # Execute the cut-short strategy.
        strategy.execute(**extra_options)

        # Repair the last reformulation.
        lastReform.successor = oldSuccessor

        # Retrieve and augment the output problem (unless it's self).
        output = lastReform.output
        if output is not self:
            output.prepared_strategy = strategy
            output.prepared_steps = steps

        return output

    def solve(self, **extra_options):
        """Hand the problem to a solver.

        You can select the solver manually with the ``solver`` option. Otherwise
        a suitable solver will be selected among those that are available on the
        platform.

        The default behavior (options ``primals=True``, ``duals=None``) is to
        raise a :exc:`~picos.SolutionFailure` when the primal solution is not
        found optimal by the solver, while the dual solution is allowed to be
        missing or incomplete.

        When this method succeeds and unless ``apply_solution=False``, you can
        access the solution as follows:

            - The problem's :attr:`value` denotes the objective function value.
            - The variables' :attr:`~picos.expressions.Expression.value` is set
              according to the primal solution. You can in fact query the value
              of any expression involving valued variables like this.
            - The constraints' :attr:`~picos.constraints.Constraint.dual` is set
              according to the dual solution.

        :param extra_options:
            A sequence of additional solver options to use with this solution
            search only. In particular, this lets you

            - select a solver via the ``solver`` option,
            - obtain non-optimal primal solutions by setting ``primals=None``,
            - require a complete and optimal dual solution with ``duals=True``,
              and
            - skip valuing variables or constraints with
              ``apply_solution=False``.

        :returns ~picos.Solution or list(~picos.Solution):
            A solution object or list thereof.

        :raises ~picos.SolutionFailure:
            In the following cases:

            1. No solution strategy was found.
            2. Multiple solutions were requested but none were returned.
            3. A primal solution was explicitly requested (``primals=True``) but
               the primal solution is missing/incomplete or not claimed optimal.
            4. A dual solution was explicitly requested (``duals=True``) but
               the dual solution is missing/incomplete or not claimed optimal.

            The case number is stored in the ``code`` attribute of the
            exception.
        """
        from .strategy import NoStrategyFound, Strategy

        startTime = time.time()

        extra_options = map_legacy_options(**extra_options)
        options = self.options.self_or_updated(**extra_options)
        verbose = options.verbosity > 0

        with picos_box(show=verbose):
            if verbose:
                print("Problem type: {}.".format(self.type))

            # Reset an outdated strategy.
            if self._strategy and not self._strategy.valid(**extra_options):
                if verbose:
                    print("Strategy outdated:\n{}.".format(self._strategy))

                self._strategy = None

            # Find a new solution strategy, if necessary.
            if not self._strategy:
                if verbose:
                    print("Searching a solution strategy{}.".format(
                    " for {}".format(get_solver(options.solver).names()[1])
                    if options.solver else ""))

                try:
                    self._strategy = Strategy.from_problem(
                        self, **extra_options)
                except NoStrategyFound as error:
                    raise SolutionFailure(1, error)

                if verbose:
                    print("Solution strategy:\n  {}".format(
                        "\n  ".join(str(self._strategy).splitlines())))
            else:
                if verbose:
                    print("Reusing strategy:\n  {}".format(
                        "\n  ".join(str(self._strategy).splitlines())))

            # Execute the strategy to obtain one or more solutions.
            solutions = self._strategy.execute(**extra_options)

            # Report how many solutions were obtained, select the first.
            if isinstance(solutions, list):
                assert all(isinstance(s, Solution) for s in solutions)

                if not solutions:
                    raise SolutionFailure(
                        2, "The solver returned an empty list of solutions.")

                solution = solutions[0]

                if verbose:
                    print("Selecting the first of {} solutions obtained for "
                        "processing.".format(len(solutions)))
            else:
                assert isinstance(solutions, Solution)
                solution = solutions

            # Report claimed solution state.
            if verbose:
                print("Solver claims {} solution for {} problem.".format(
                    solution.claimedStatus, solution.problemStatus))

            # Validate the primal solution.
            if options.primals:
                if solution.primalStatus != SS_OPTIMAL:
                    raise SolutionFailure(3, "Primal solution state claimed {} "
                        "but optimality is required (primals=True).".format(
                        solution.primalStatus))
                elif None in solution.primals:
                    raise SolutionFailure(
                        "The primal solution is incomplete (primals=True).")

            # Validate the dual solution.
            if options.duals:
                if solution.dualStatus != SS_OPTIMAL:
                    raise SolutionFailure(4, "Dual solution state claimed {} "
                        "but optimality is required (duals=True).".format(
                        solution.dualStatus))
                elif None in solution.duals:
                    raise SolutionFailure(
                        "The dual solution is incomplete (duals=True).")

            if options.apply_solution:
                if verbose:
                    print("Applying the solution.")

                # Apply the (first) solution.
                solution.apply(snapshotStatus=True)

                # Store all solutions produced by the solver.
                self._last_solution = solutions

                # Report verified solution state.
                if verbose:
                    print("Applied solution is {}.".format(solution.lastStatus))

            endTime = time.time()
            solveTime = endTime - startTime
            searchTime = solution.searchTime

            if searchTime:
                overhead = (solveTime - searchTime) / searchTime
            else:
                overhead = float("inf")

            if verbose:
                print("Search {:.1e}s, solve {:.1e}s, overhead {:.0%}."
                    .format(searchTime, solveTime, overhead))

        return solutions

    @deprecated("2.0", reason="Misleading semantics. Maybe "
        ":func:`picos.minimize` is what you want.")
    def minimize(self, obj, **extra_options):
        """Look for a minimizing solution.

        Sets the objective to minimize the given objective function and calls
        the solver with the given additional options.

        :param obj: The objective function to minimize.
        :type obj: :class:`~picos.expressions.Expression`
        :param extra_options: A sequence of additional solver options.

        :returns: A dictionary, see :func:`solve <picos.Problem.solve>`.

        .. warning::
            This is equivalent to
            :func:`set_objective <picos.Problem.set_objective>`
            followed by :func:`solve <picos.Problem.solve>` and will thus
            override any existing objective function and direction.
        """
        self.objective = "min", obj
        return self.solve(**extra_options)

    @deprecated("2.0", reason="Misleading semantics. Maybe "
        ":func:`picos.maximize` is what you want.")
    def maximize(self, obj, **extra_options):
        """Look for a maximization solution.

        Sets the objective to maximize the given objective function and calls
        the solver with the given additional options.

        :param obj: The objective function to maximize.
        :type obj: :class:`~picos.expressions.Expression`
        :param extra_options: A sequence of additional solver options.

        :returns: A dictionary, see :func:`solve <picos.Problem.solve>`.

        .. warning::
            This is equivalent to
            :func:`set_objective <picos.Problem.set_objective>`
            followed by :func:`solve <picos.Problem.solve>` and will thus
            override any existing objective function and direction.
        """
        self.objective = "max", obj
        return self.solve(**extra_options)

    def write_to_file(self, filename, writer="picos"):
        """See :func:`picos.modeling.file_out.write`."""
        write(self, filename, writer)

    # --------------------------------------------------------------------------
    # Methods to query the problem.
    # TODO: Document removal of is_complex, is_real (also for constraints).
    # TODO: Revisit #14: "Interfaces to get primal/dual objective values and
    #       primal/dual feasiblity (amount of violation).""
    # --------------------------------------------------------------------------

    def check_current_value_feasibility(self, tol=1e-5, inttol=None):
        """Check if the problem is feasibly valued.

        Checks whether all variables that appear in constraints are valued and
        satisfy both their bounds and the constraints up to the given tolerance.

        :param float tol: Largest tolerated absolute violation of a constraint
            or variable bound. If ``None``, then the ``abs_prim_fsb_tol`` solver
            option is used.
        :param inttol: DEPRECATED

        :returns: A tuple ``(feasible, violation)`` where ``feasible`` is a bool
            stating whether the solution is feasible and ``violation`` is either
            ``None``, if ``feasible == True``, or the amount of violation,
            otherwise.
        """
        if inttol is not None:
            throw_deprecation_warning("Variable integrality is now ensured on "
                "assignment of a value, so it does not need to be checked via "
                "check_current_value_feasibility's old 'inttol' parameter.")

        if tol is None:
            tol = self._options.abs_prim_fsb_tol

        all_cons = list(self._constraints.values())
        all_cons += [
            variable.bound_constraint for variable in self.variables.values()
            if variable.bound_constraint]

        largest_violation = 0.0

        for constraint in all_cons:
            slack = constraint.slack

            assert isinstance(slack, (float, cvx.matrix, cvx.spmatrix))
            if isinstance(slack, (float, cvx.spmatrix)):
                slack = cvx.matrix(slack)  # Allow min, max.

            # TODO: Add a "violation" interface to constraints?
            if isinstance(constraint, constraints.LMIConstraint):
                # Check hermitian-ness of slack.
                violation = float(max(abs(slack - slack.H)))
                if violation > tol:
                    largest_violation = max(largest_violation, violation)

                # Check positive semidefiniteness of slack.
                violation = -float(min(np.linalg.eigvalsh(cvx2np(slack))))
                if violation > tol:
                    largest_violation = max(largest_violation, violation)
            else:
                violation = -float(min(slack))
                if violation > tol:
                    largest_violation = max(largest_violation, violation)

        return (not largest_violation, largest_violation)

    # --------------------------------------------------------------------------
    # Legacy methods and properties.
    # --------------------------------------------------------------------------

    _LEGACY_PROPERTY_REASON = "Still used internally by legacy code; will be " \
        "removed together with that code."

    @property
    @deprecated("2.0", reason=_LEGACY_PROPERTY_REASON)
    def countVar(self):
        """The same as :func:`len` applied to :attr:`variables`."""
        return len(self.variables)

    @property
    @deprecated("2.0", reason=_LEGACY_PROPERTY_REASON)
    def countCons(self):
        """The same as :func:`len` applied to :attr:`constraints`."""
        return len(self.variables)

    @property
    @deprecated("2.0", reason=_LEGACY_PROPERTY_REASON)
    def numberOfVars(self):
        """The sum of the dimensions of all referenced variables."""
        return sum(variable.dim for variable in self.variables.values())

    @property
    @deprecated("2.0", reason=_LEGACY_PROPERTY_REASON)
    def numberLSEConstraints(self):
        """Number of :class:`~picos.constraints.LogSumExpConstraint` stored."""
        return len([c for c in self._constraints.values()
            if isinstance(c, constraints.LogSumExpConstraint)])

    @property
    @deprecated("2.0", reason=_LEGACY_PROPERTY_REASON)
    def numberSDPConstraints(self):
        """Number of :class:`~picos.constraints.LMIConstraint` stored."""
        return len([c for c in self._constraints.values()
            if isinstance(c, constraints.LMIConstraint)])

    @property
    @deprecated("2.0", reason=_LEGACY_PROPERTY_REASON)
    def numberQuadConstraints(self):
        """Number of quadratic constraints stored."""
        return len([c for c in self._constraints.values() if isinstance(c, (
            constraints.ConvexQuadraticConstraint,
            constraints.ConicQuadraticConstraint,
            constraints.NonconvexQuadraticConstraint))])

    @property
    @deprecated("2.0", reason=_LEGACY_PROPERTY_REASON)
    def numberConeConstraints(self):
        """Number of quadratic conic constraints stored."""
        return len([c for c in self._constraints.values() if isinstance(
            c, (constraints.SOCConstraint, constraints.RSOCConstraint))])

    @deprecated("2.0", useInstead="value")
    def obj_value(self):
        """Objective function value.

        :raises AttributeError:
            If the problem is a feasibility problem or if the objective function
            is not valued. This is legacy behavior. Note that :attr:`value` just
            returns :obj:`None` while functions that **do** raise an exception
            to denote an unvalued expression would raise
            :exc:`~picos.expressions.NotValued` instead.
        """
        if self._objective.feasibility:
            raise AttributeError(
                "A feasibility problem has no objective value.")

        value = self.value

        if self.value is None:
            raise AttributeError("The objective {} is not fully valued."
                .format(self._objective.function.string))
        else:
            return value

    @deprecated("2.0", useInstead="continuous")
    def is_continuous(self):
        """Whether all variables are of continuous types."""
        return self.continuous

    @deprecated("2.0", useInstead="pure_integer")
    def is_pure_integer(self):
        """Whether all variables are of integral types."""
        return self.pure_integer

    @deprecated("2.0", useInstead="Problem.options")
    def set_all_options_to_default(self):
        """Set all solver options to their default value."""
        self._options.reset()

    @deprecated("2.0", useInstead="Problem.options")
    def set_option(self, key, val):
        """Set a single solver option to the given value.

        :param str key: String name of the option, see below for a list.
        :param val: New value for the option.
        """
        key, val = map_legacy_options({key: val}).popitem()
        self._options[key] = val

    @deprecated("2.0", useInstead="Problem.options")
    def update_options(self, **options):
        """Set multiple solver options at once.

        :param options: A parameter sequence of options to set.
        """
        options = map_legacy_options(**options)
        for key, val in options.items():
            self._options[key] = val

    @deprecated("2.0", useInstead="Problem.options")
    def verbosity(self):
        """Return the problem's current verbosity level."""
        return self._options.verbosity

    @deprecated("2.0", reason="Variables can now be created independent of "
        "problems, and do not need to be added to any problem explicitly.")
    def add_variable(
            self, name, size=1, vtype='continuous', lower=None, upper=None):
        r"""Legacy method to create a PICOS variable.

        :param str name: The name of the variable.

        :param size:
            The shape of the variable.
        :type size:
            anything recognized by :func:`~picos.expressions.data.load_shape`

        :param str vtype:
            Domain of the variable. Can be any of

            - ``'continuous'`` -- real valued,
            - ``'binary'`` -- either zero or one,
            - ``'integer'`` -- integer valued,
            - ``'symmetric'`` -- symmetric matrix,
            - ``'antisym'`` or ``'skewsym'`` -- skew-symmetric matrix,
            - ``'complex'`` -- complex matrix,
            - ``'hermitian'`` -- complex hermitian matrix.

        :param lower:
            A lower bound on the variable.
        :type lower:
            anything recognized by :func:`~picos.expressions.data.load_data`

        :param upper:
            An upper bound on the variable.
        :type upper:
            anything recognized by :func:`~picos.expressions.data.load_data`

        :returns:
            A :class:`~picos.expressions.BaseVariable` instance.

        :Example:

        >>> from picos import Problem
        >>> P = Problem()
        >>> x = P.add_variable("x", 3)
        >>> x
        <3×1 Real Variable: x>
        >>> # Variable are not stored inside the problem any more:
        >>> P.variables
        mappingproxy(OrderedDict())
        >>> # They are only part of the problem if they actually appear:
        >>> P.set_objective("min", abs(x)**2)
        >>> P.variables
        mappingproxy(OrderedDict([('x', <3×1 Real Variable: x>)]))
        """
        if vtype == "continuous":
            return expressions.RealVariable(name, size, lower, upper)
        elif vtype == "binary":
            return expressions.BinaryVariable(name, size)
        elif vtype == "integer":
            return expressions.IntegerVariable(name, size, lower, upper)
        elif vtype == "symmetric":
            return expressions.SymmetricVariable(name, size, lower, upper)
        elif vtype in ("antisym", "skewsym"):
            return expressions.SkewSymmetricVariable(name, size, lower, upper)
        elif vtype == "complex":
            return expressions.ComplexVariable(name, size)
        elif vtype == "hermitian":
            return expressions.HermitianVariable(name, size)
        elif vtype in ("semiint", "semicont"):
            raise NotImplementedError("Variables with legacy types 'semiint' "
                "and 'semicont' are not supported anymore as of PICOS 2.0. "
                "If you need this functionality back, please open an issue.")
        else:
            raise ValueError("Unknown legacy variable type '{}'.".format(vtype))

    @deprecated("2.0", reason="Whether a problem references a variable is now"
        " determined dynamically, so this method has no effect.")
    def remove_variable(self, name):
        """Does nothing."""
        pass

    @deprecated("2.0", useInstead="variables")
    def set_var_value(self, name, value):
        """Set the :attr:`~picos.expressions.Expression.value` of a variable.

        For a :class:`Problem` ``P``, this is the same as
        ``P.variables[name] = value``.

        :param str name:
            Name of the variable to be valued.

        :param value:
            The value to be set.
        :type value:
            anything recognized by :func:`~picos.expressions.data.load_data`
        """
        try:
            variable = self.variables[name]
        except KeyError:
            raise KeyError(
                "The problem references no variable named '{}'.".format(name))
        else:
            variable.value = value

    @deprecated("2.0", useInstead="dual")
    def as_dual(self):
        """Return the Lagrangian dual problem of the standardized problem."""
        return self.dual


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements :class:`SecondOrderCone`."""

import operator
from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from ..constraints import SOCConstraint
from .data import convert_operands
from .exp_affine import AffineExpression
from .expression import refine_operands, validate_prediction
from .set import Set

_API_START = api_start(globals())
# -------------------------------


class SecondOrderCone(Set):
    r"""The second order cone.

    .. _lorentz:

    Also known as the quadratic, :math:`2`-norm, Lorentz, or ice cream cone.

    For :math:`n \in \mathbb{Z}_{\geq 2}`, represents the convex cone

    .. math::

        \mathcal{Q}^n = \left\{
            x \in \mathbb{R}^n
        ~\middle|~
            x_1 \geq \sqrt{\sum_{i = 2}^n x_i^2}
        \right\}.

    :Dual cone:

    The second order cone as defined above is self-dual.
    """

    def __init__(self):
        """Construct a second order cone."""
        typeStr = "Second Order Cone"
        symbStr = glyphs.set(glyphs.sep(
            glyphs.col_vectorize("t", "x"), glyphs.le(glyphs.norm("x"), "t")))

        Set.__init__(self, typeStr, symbStr)

    def _get_variables(self):
        return set()

    def _replace_variables(self):
        return self

    Subtype = namedtuple("Subtype", ())

    def _get_subtype(self):
        return self.Subtype()

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        if relation == operator.__rshift__:
            if issubclass(other.clstype, AffineExpression):
                if other.subtype.dim >= 2:
                    return SOCConstraint.make_type(other.subtype.dim - 1)

        return NotImplemented

    @convert_operands()
    @validate_prediction
    @refine_operands()
    def __rshift__(self, element):
        if isinstance(element, AffineExpression):
            if len(element) < 2:
                raise TypeError("Elements of the second order cone must be "
                    "at least two-dimensional.")

            element = element.vec

            return SOCConstraint(element[1:], element[0])
        else:
            return NotImplemented


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2012-2017 Guillaume Sagnol
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`DetRootNConstraint`."""

from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class DetRootNConstraint(Constraint):
    """Lower bound on the :math:`n`-th root of a matrix determinant."""

    class Conversion(ConstraintConversion):
        """:math:`n`-th root of a matrix determinant constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from ..expressions import LowerTriangularVariable
            from . import (ComplexLMIConstraint, GeometricMeanConstraint,
                           LMIConstraint)

            n = subtype.diag
            r = (n * (n + 1)) // 2

            yield (
                "var", LowerTriangularVariable.make_var_type(dim=r, bnd=0), 1)

            if subtype.complex:
                yield ("con", ComplexLMIConstraint.make_type(diag=2*n), 1)
            else:
                yield ("con", LMIConstraint.make_type(diag=2*n), 1)

            yield ("con", GeometricMeanConstraint.make_type(argdim=n), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..modeling import Problem
            from ..expressions import GeometricMean, LowerTriangularVariable

            n = con.detRootN.n

            P = Problem()
            L = LowerTriangularVariable("__L", n)
            d = L.maindiag
            D = d.diag
            P.add_constraint((con.detRootN.x & L) // (L.T & D) >> 0)
            P.add_constraint(GeometricMean(d) >= con.lowerBound)
            return P

    def __init__(self, detRootN, lowerBound):
        """Construct a :class:`DetRootNConstraint`.

        :param ~picos.expressions.DetRootN detRootN:
            Constrained expression.
        :param ~picos.expressions.AffineExpression lowerBound:
            Lower bound on the expression.
        """
        from ..expressions import AffineExpression, DetRootN

        assert isinstance(detRootN, DetRootN)
        assert isinstance(lowerBound, AffineExpression)
        assert len(lowerBound) == 1

        self.detRootN   = detRootN
        self.lowerBound = lowerBound

        super(DetRootNConstraint, self).__init__(detRootN._typeStr)

    Subtype = namedtuple("Subtype", ("diag", "complex"))

    def _subtype(self):
        return self.Subtype(self.detRootN.n, self.detRootN.x.complex)

    @classmethod
    def _cost(cls, subtype):
        n = subtype.diag

        if subtype.complex:
            return n**2 + 1
        else:
            return n*(n + 1)//2 + 1

    def _expression_names(self):
        yield "detRootN"
        yield "lowerBound"

    def _str(self):
        return glyphs.ge(self.detRootN.string, self.lowerBound.string)

    def _get_slack(self):
        return self.detRootN.value - self.lowerBound.value


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2012-2017 Guillaume Sagnol
# Copyright (C) 2017-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`CVXOPTSolver`."""

import cvxopt
import numpy

from ..apidoc import api_end, api_start
from ..constraints import (AffineConstraint, LMIConstraint,
                           LogSumExpConstraint, RSOCConstraint, SOCConstraint)
from ..expressions import CONTINUOUS_VARTYPES, AffineExpression, LogSumExp
from ..modeling.footprint import Specification
from ..modeling.solution import (PS_FEASIBLE, PS_INFEASIBLE, PS_UNBOUNDED,
                                 PS_UNKNOWN, SS_INFEASIBLE, SS_OPTIMAL,
                                 SS_UNKNOWN)
from .solver import Solver, SolverError

_API_START = api_start(globals())
# -------------------------------


class CVXOPTSolver(Solver):
    """Interface to the CVXOPT solver.

    Also used as an interface to the
    :class:`SMCP solver <picos.solvers.solver_smcp.SMCPSolver>`.
    """

    SUPPORTED = Specification.compile(
        directions=Specification.ALL,
        objectives=[
            AffineExpression,
            LogSumExp],
        variables=CONTINUOUS_VARTYPES,
        constraints=[
            AffineConstraint,
            SOCConstraint,
            RSOCConstraint,
            LMIConstraint,
            LogSumExpConstraint])

    @classmethod
    def supports(cls, footprint):
        """Implement :meth:`~.solver.Solver.supports`."""
        if not Solver.supports(footprint):
            return False

        return footprint << cls.SUPPORTED

    @classmethod
    def default_penalty(cls):
        """Implement :meth:`~.solver.Solver.default_penalty`."""
        return 1.0  # Stable free/open source solver.

    @classmethod
    def test_availability(cls):
        """Implement :meth:`~.solver.Solver.test_availability`."""
        # CVXOPT is a dependency of PICOS, so it is always available.
        pass

    @classmethod
    def names(cls):
        """Implement :meth:`~.solver.Solver.names`."""
        return "cvxopt", "CVXOPT", "Python Convex Optimization Solver"

    @property
    def is_smcp(self):
        """Whether to implement SMCP instead of CVXOPT."""
        return False

    def __init__(self, problem):
        """Initialize a CVXOPT solver interface.

        :param ~picos.Problem problem: The problem to be solved.
        """
        super(CVXOPTSolver, self).__init__(problem)

        self._numVars = 0
        """Total number of scalar variables passed to CVXOPT."""

        self._cvxoptVarOffset = {}
        """Maps a PICOS variable to its offset in the constraint matrix."""

        # HACK: Setting this to false would result in variable bounds to be
        #       ignored, instead of added to the linear inequalities matrix.
        #       This is used by a Problem method that prints the problem to a
        #       file using a CVXOPTSolver instance.
        self.import_variable_bounds = True

        # SMCP currently offers no option reset, so keep a backup.
        if self.is_smcp:
            import smcp
            self._smcp_default_options = smcp.solvers.options.copy()

    def reset_problem(self):
        """Implement :meth:`~.solver.Solver.reset_problem`."""
        self.int = None

        self._numVars = 0
        self._cvxoptVarOffset.clear()

    def _is_geometric_program(self):
        return isinstance(self.ext.no.function, LogSumExp) or \
            any(isinstance(constraint, LogSumExpConstraint)
            for constraint in self.ext.constraints.values())

    def _affine_expression_to_G_and_h(self, expression):
        assert isinstance(expression, AffineExpression)

        length = len(expression)

        # Construct G.
        I, J, V = [], [], []
        for variable, coefficients in expression._coefs.items():
            cvxoptVarOffset = self._cvxoptVarOffset[variable]

            if not isinstance(coefficients, cvxopt.spmatrix):
                coefficients = cvxopt.sparse(coefficients)

            I.extend(coefficients.I)
            J.extend([cvxoptVarOffset + j for j in coefficients.J])
            V.extend(coefficients.V)
        G = cvxopt.spmatrix(V, I, J, (length, self._numVars), tc="d")

        # Construct h.
        # TODO: Check if making a copy is necessary.
        h = cvxopt.matrix(expression._const, tc="d")

        return G, h

    _Gh = _affine_expression_to_G_and_h

    def _import_variables_without_bounds(self):
        offset = 0
        for variable in self.ext.variables.values():
            dim = variable.dim

            # Register the variable.
            self._cvxoptVarOffset[variable] = offset
            offset += dim

        assert offset == self._numVars

    def _import_variable_bounds(self):
        for variable in self.ext.variables.values():
            if self.import_variable_bounds:
                bounds = variable.bound_constraint
                if bounds:
                    self._import_affine_constraint(bounds)

    def _import_objective(self):
        direction, objective = self.ext.no

        if self._is_geometric_program():
            if isinstance(objective, LogSumExp):
                (F, g) = self._Gh(objective.x)
            else:
                assert isinstance(objective, AffineExpression)
                (F, g) = self._Gh(objective)

            # NOTE: Needs to happen before LogSumExpConstraint are added.
            self.int["K"] = [F.size[0]]

            if direction == "max":
                F, g = -F, -g

            self.int["F"] = F
            self.int["g"] = g
        else:
            (c, _) = self._Gh(objective)

            if direction == "max":
                c = -c

            # Must be a dense column-vector.
            c = cvxopt.matrix(c).T

            self.int["c"] = c

    def _import_affine_constraint(self, constraint):
        assert isinstance(constraint, AffineConstraint)

        (G_smaller, h_smaller) = self._Gh(constraint.smaller)
        (G_greater, h_greater) = self._Gh(constraint.greater)

        G = G_smaller - G_greater
        h = h_greater - h_smaller

        if constraint.is_equality():
            self.int["A"] = cvxopt.sparse([self.int["A"], G])
            self.int["b"] = cvxopt.matrix([self.int["b"], h])
        else:
            self.int["Gl"] = cvxopt.sparse([self.int["Gl"], G])
            self.int["hl"] = cvxopt.matrix([self.int["hl"], h])

    def _import_soc_constraint(self, constraint):
        assert isinstance(constraint, SOCConstraint)

        (A, b) = self._Gh(constraint.ne)
        (c, d) = self._Gh(constraint.ub)

        self.int["Gq"].append(cvxopt.sparse([-c, -A]))
        self.int["hq"].append(cvxopt.matrix([d, b]))

    def _import_rsoc_constraint(self, constraint):
        assert isinstance(constraint, RSOCConstraint)

        (A,  b)  = self._Gh(constraint.ne)
        (c1, d1) = self._Gh(constraint.ub1)
        (c2, d2) = self._Gh(constraint.ub2)

        self.int["Gq"].append(cvxopt.sparse([-c1 - c2, -2 * A, c2 - c1]))
        self.int["hq"].append(cvxopt.matrix([d1 + d2, 2 * b, d1 - d2]))

    def _import_lmi_constraint(self, constraint):
        assert isinstance(constraint, LMIConstraint)

        (G_smaller, h_smaller) = self._Gh(constraint.smaller)
        (G_greater, h_greater) = self._Gh(constraint.greater)

        self.int["Gs"].append(G_smaller - G_greater)
        self.int["hs"].append(h_greater - h_smaller)

    def _import_lse_constraint(self, constraint):
        assert isinstance(constraint, LogSumExpConstraint)

        (F, g) = self._Gh(constraint.le0.x)

        self.int["F"] = cvxopt.sparse([self.int["F"], F])
        self.int["g"] = cvxopt.matrix([self.int["g"], g])
        self.int["K"].append(F.size[0])

    def _import_constraint(self, constraint):
        if isinstance(constraint, AffineConstraint):
            self._import_affine_constraint(constraint)
        elif isinstance(constraint, SOCConstraint):
            self._import_soc_constraint(constraint)
        elif isinstance(constraint, RSOCConstraint):
            self._import_rsoc_constraint(constraint)
        elif isinstance(constraint, LMIConstraint):
            self._import_lmi_constraint(constraint)
        elif isinstance(constraint, LogSumExpConstraint):
            self._import_lse_constraint(constraint)
        else:
            assert False, "Unexpected constraint type."

    def _import_problem(self):
        self._numVars = sum(var.dim for var in self.ext.variables.values())

        # CVXOPT's internal problem representation is stateless; a number of
        # matrices are supplied to the appropriate solver function each time a
        # search is started. These matrices are thus stored in self.int.
        self.int = {
            # Objective function coefficients.
            "c": None,

            # Linear equality left hand side.
            "A": cvxopt.spmatrix([], [], [], (0, self._numVars), tc="d"),

            # Linear equality right hand side.
            "b": cvxopt.matrix([], (0, 1), tc="d"),

            # Linear inequality left hand side.
            "Gl": cvxopt.spmatrix([], [], [], (0, self._numVars), tc="d"),

            # Linear inequality right hand side.
            "hl": cvxopt.matrix([], (0, 1), tc="d"),

            # Second order cone inequalities left hand sides.
            "Gq": [],

            # Second order cone inequalities right hand sides.
            "hq": [],

            # Semidefinite cone inequalities left hand sides.
            "Gs": [],

            # Semidefinite cone inequalities right hand sides.
            "hs": [],

            # Geometric program data.
            "F": None,
            "g": None,
            "K": None
        }

        # Import variables without their bounds.
        self._import_variables_without_bounds()

        # Set objective.
        # NOTE: This needs to happen before constraints are added as
        #       self.int["K"][0] refers to an LSE objective while
        #       self.int["K"][i] with i > 0 refers to LSE constraints.
        self._import_objective()

        # Import constraints.
        for constraint in self.ext.constraints.values():
            self._import_constraint(constraint)

        # Import variable bounds as additional affine constraints.
        # NOTE: This needs to happen after constraints are added due to how the
        #       dual values for constraints are extracted.
        self._import_variable_bounds()

    def _update_problem(self):
        raise NotImplementedError

    def _solve(self):
        if self.is_smcp:
            import smcp

        p = self.int
        isGP = self._is_geometric_program()

        # Clear all options set previously. This is necessary because CVXOPT
        # options are global, and might be changed even by another problem.
        cvxopt.solvers.options.clear()

        # verbosity
        cvxopt.solvers.options["show_progress"] = (self.verbosity() >= 1)

        # rel_prim_fsb_tol, rel_dual_fsb_tol
        feasibilityTols = [tol for tol in (self.ext.options.rel_prim_fsb_tol,
                self.ext.options.rel_dual_fsb_tol) if tol is not None]
        if feasibilityTols:
            cvxopt.solvers.options["feastol"] = min(feasibilityTols)

        # abs_ipm_opt_tol
        if self.ext.options.abs_ipm_opt_tol is not None:
            cvxopt.solvers.options["abstol"] = self.ext.options.abs_ipm_opt_tol

        # rel_ipm_opt_tol
        if self.ext.options.rel_ipm_opt_tol is not None:
            cvxopt.solvers.options["reltol"] = self.ext.options.rel_ipm_opt_tol

        # max_iterations
        if self.ext.options.max_iterations is not None:
            cvxopt.solvers.options["maxiters"] = self.ext.options.max_iterations
        else:
            cvxopt.solvers.options["maxiters"] = int(1e6)

        # Allow solving GPs that are not strictly feasible and work around
        # constraint matrix rank requirements.
        if self.ext.options.cvxopt_kktreg is not None:
            cvxopt.solvers.options["kktreg"] = self.ext.options.cvxopt_kktreg

        # Handle unsupported options.
        self._handle_unsupported_options(
            "lp_root_method", "lp_node_method", "timelimit", "treememory",
            "max_fsb_nodes", "hotstart")

        # TODO: Add CVXOPT-sepcific options. Candidates are:
        #       - refinement

        # Set options for SMCP.
        if self.is_smcp:
            # Restore default options.
            smcp.solvers.options = self._smcp_default_options.copy()

            # Copy options also used by CVXOPT.
            smcp.solvers.options.update(cvxopt.solvers.options)

            # Further handle "verbose" option.
            smcp.solvers.options["debug"] = (self.verbosity() >= 2)

            # TODO: Add SMCP-sepcific options.

        if self._debug():
            from pprint import pformat
            self._debug("Setting options:\n{}\n".format(pformat(
                smcp.solvers.options if self.is_smcp
                else cvxopt.solvers.options)))

        # Print a header.
        if self.is_smcp:
            subsolverText = None
        else:
            if isGP:
                subsolverText = "internal GP solver"
            else:
                subsolverText = "internal CONELP solver"

        # Further prepare the problem for the CVXOPT/SMCP CONELP solvers.
        # TODO: This should be done during import.
        if not isGP:
            # Retrieve the structure of the cone, which is a cartesian product
            # of the non-negative orthant of dimension l, a number of second
            # order cones with dimensions in q and a number of positive
            # semidefinite cones with dimensions in s.
            dims = {
                "l": p["Gl"].size[0],
                "q": [Gqi.size[0] for Gqi in p["Gq"]],
                "s": [int(numpy.sqrt(Gsi.size[0])) for Gsi in p["Gs"]]
            }

            # Construct G and h to contain all conic inequalities, starting with
            # those with respect to the non-negative orthant.
            G = p["Gl"]
            h = p["hl"]

            # SMCP's ConeLP solver does not handle (linear) equalities, so cast
            # them as inequalities.
            if self.is_smcp:
                smcp_eps = min(feasibilityTols)
                if p["A"].size[0] > 0:
                    G = cvxopt.sparse([G, p["A"]])
                    G = cvxopt.sparse([G, -p["A"]])
                    h = cvxopt.matrix([h, p["b"]+smcp_eps])
                    h = cvxopt.matrix([h, smcp_eps-p["b"]])
                    dims["l"] += (2 * p["A"].size[0])

                # Remove the lines in G and h corresponding to 0==0 or 0<=0
                JP = list(set(G.I))
                IP = range(len(JP))
                VP = [1] * len(JP)

                if len(JP) != dims["l"]:
                    # is there a constraint of the form 0<=a, (a<0) ?
                    if any([b < -smcp_eps for (i, b) in enumerate(h)
                            if i not in JP]):
                        raise Exception(
                            'infeasible constraint of the form '
                            '0 <= a, with a<0')

                    # left-multiply with PPP-matrix to remove 0-constraints
                    PPP = cvxopt.spmatrix(VP, IP, JP, (len(IP), G.size[0]))
                    dims["l"] = len(JP)
                    G = PPP * G
                    h = PPP * h

            # Add second-order cone inequalities.
            for i in range(len(dims["q"])):
                G = cvxopt.sparse([G, p["Gq"][i]])
                h = cvxopt.matrix([h, p["hq"][i]])

            # Add semidefinite cone inequalities.
            for i in range(len(dims["s"])):
                G = cvxopt.sparse([G, p["Gs"][i]])
                h = cvxopt.matrix([h, p["hs"][i]])

            # Remove zero lines from linear equality constraint matrix, as
            # CVXOPT expects this matrix to have full row rank.
            JP = list(set(p["A"].I))
            IP = range(len(JP))
            VP = [1] * len(JP)
            if any([b for (i, b) in enumerate(p["b"]) if i not in JP]):
                raise SolverError(
                    "Infeasible constraint of the form 0 = a with a != 0.")
            P = cvxopt.spmatrix(VP, IP, JP, (len(IP), p["A"].size[0]))
            A = P * p["A"]
            b = P * p["b"]

        # Attempt to solve the problem.
        with self._header(subsolverText), self._stopwatch():
            if self.is_smcp:
                if self._debug():
                    self._debug("Calling smcp.solvers.conelp(c, G, h, dims) "
                        "with\nc:\n{}\nG:\n{}\nh:\n{}\ndims:\n{}\n"
                        .format(p["c"], G, h, dims))
                try:
                    result = smcp.solvers.conelp(p["c"], G, h, dims)
                except TypeError:
                    # HACK: Work around "'NoneType' object is not subscriptable"
                    #       exception with infeasible/unbounded problems.
                    result = None
            else:
                kwargs = {}
                if self.ext.options.cvxopt_kktsolver is not None:
                    kwargs["kktsolver"] = self.ext.options.cvxopt_kktsolver

                if isGP:
                    result = cvxopt.solvers.gp(p["K"], p["F"], p["g"], p["Gl"],
                        p["hl"], p["A"], p["b"], **kwargs)
                else:
                    result = cvxopt.solvers.conelp(
                        p["c"], G, h, dims, A, b, **kwargs)

        # Retrieve primals.
        primals = {}
        if self.ext.options.primals is not False and result is not None \
        and result["x"] is not None:
            for variable in self.ext.variables.values():
                offset = self._cvxoptVarOffset[variable]
                value = list(result["x"][offset:offset + variable.dim])
                primals[variable] = value

        # Retrieve duals.
        duals = {}
        if self.ext.options.duals is not False and result is not None:
            (indy, indzl, indzq, indznl, indzs) = (0, 0, 0, 0, 0)

            if isGP:
                zkey  = "zl"
                zqkey = "zq"
                zskey = "zs"
            else:
                zkey  = "z"
                zqkey = "z"
                zskey = "z"
                indzq = dims["l"]
                indzs = dims["l"] + sum(dims["q"])

            if self.is_smcp:
                # Equality constraints were cast as two inequalities.
                ieq = p["Gl"].size[0]
                neq = (dims["l"] - ieq) // 2
                soleq = result["z"][ieq:ieq + neq]
                soleq -= result["z"][ieq + neq:ieq + 2 * neq]
            else:
                soleq = result["y"]

            for constraint in self.ext.constraints.values():
                dual   = None
                consSz = len(constraint)

                if isinstance(constraint, AffineConstraint):
                    if constraint.is_equality():
                        if soleq is not None:
                            dual = (P.T * soleq)[indy:indy + consSz]
                            indy += consSz
                    else:
                        if result[zkey] is not None:
                            dual = result[zkey][indzl:indzl + consSz]
                            indzl += consSz
                elif isinstance(constraint, SOCConstraint) \
                or isinstance(constraint, RSOCConstraint):
                    if result[zqkey] is not None:
                        if isGP:
                            dual = result[zqkey][indzq]
                            dual[1:] = -dual[1:]
                            indzq += 1
                        else:
                            dual = result[zqkey][indzq:indzq + consSz]
                            if isinstance(constraint, RSOCConstraint):
                                # RScone were cast as a SOcone on import, so
                                # transform the dual to a proper RScone dual.
                                alpha = dual[0] + dual[-1]
                                beta  = dual[0] - dual[-1]
                                z     = 2.0 * dual[1:-1]
                                dual  = cvxopt.matrix([alpha, beta, z])
                            indzq += consSz
                elif isinstance(constraint, LMIConstraint):
                    if result[zskey] is not None:
                        matSz = constraint.size[0]
                        if isGP:
                            dual = cvxopt.matrix(
                                result[zskey][indzs], (matSz, matSz))
                            indzs += 1
                        else:
                            dual = cvxopt.matrix(
                                result[zskey][indzs:indzs + consSz],
                                (matSz, matSz))
                            indzs += consSz
                elif isinstance(constraint, LogSumExpConstraint):
                    # TODO: Retrieve LSE duals.
                    indznl += 1

                duals[constraint] = dual

        status = result["status"] if result else "unknown"
        if   status == "optimal":
            primalStatus   = SS_OPTIMAL
            dualStatus     = SS_OPTIMAL
            problemStatus  = PS_FEASIBLE
        elif status == "primal infeasible":
            primalStatus   = SS_INFEASIBLE
            dualStatus     = SS_UNKNOWN
            problemStatus  = PS_INFEASIBLE
        elif status == "dual infeasible":
            primalStatus   = SS_UNKNOWN
            dualStatus     = SS_INFEASIBLE
            problemStatus  = PS_UNBOUNDED
        else:
            primalStatus   = SS_UNKNOWN
            dualStatus     = SS_UNKNOWN
            problemStatus  = PS_UNKNOWN

        return self._make_solution(primals, duals, primalStatus, dualStatus,
            problemStatus, {"cvxopt_sol": result})


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Python 2/3 compatibility helpers."""

import sys
from abc import ABCMeta
from abc import abstractmethod as _abstractmethod
from copy import copy

from .apidoc import api_end, api_start

_API_START = api_start(globals())
# -------------------------------


VERSION = sys.version_info  #: The same as :obj:`sys.version_info`.
MAJOR   = VERSION.major  #: Major Python version.
MINOR   = VERSION.minor  #: Minor Python version.

# Import the 'builtins' module for both Pythons.
if MAJOR >= 3:
    import builtins  # noqa
else:
    import __builtin__ as builtins  # noqa

# Import 'MutableSet' for both Pythons.
if MAJOR >= 3:
    from collections.abc import MutableSet  # noqa
else:
    from collections import MutableSet  # noqa

# Allow the long type to be referenced in Python 3.
long = int if MAJOR >= 3 else long  # noqa

# Import abstract base class support for both Pythons.
ABC = ABCMeta("ABC", (object,), {"__slots__": ()})
ABC.__doc__ = "Base class for abstract classes."
abstractmethod = copy(_abstractmethod)
# HACK: Fix abstractmethod's docstring so Sphinx doesn't warn about it.
abstractmethod.__doc__ = "A decorator indicating abstract methods."


# TODO: Implement lru_cache for Python 2.
def _lru_cache(maxsize=128, typed=False):
    def decorator(function):
        def wrapper(*args, **kwargs):
            return function(*args, **kwargs)
        return wrapper
    return decorator


# Make lru_cache available.
try:
    from functools import lru_cache
except ImportError:
    try:
        from functools32 import lru_cache
    except ImportError:
        try:
            from repoze.lru import lru_cache
        except ImportError:
            lru_cache = _lru_cache


# Import or implement MappingProxyType.
try:
    from types import MappingProxyType
except ImportError:  # Python < 3.3
    from collections import Mapping as _Mapping

    class MappingProxyType(_Mapping):  # noqa
        def __init__(self, mapping):  # noqa
            self._mapping = mapping

        def __getitem__(self, key):
            return self._mapping[key]

        def __iter__(self):
            return iter(self._mapping)

        def __len__(self):
            return len(self._mapping)

        def copy(self):  # noqa
            return self._mapping.copy()


def import_exists(importName):
    """Check if a package or module name is available for import."""
    if MAJOR >= 3:
        if MINOR >= 4:
            import importlib.util
            return importlib.util.find_spec(importName) is not None
        else:
            import importlib
            return importlib.find_loader(importName) is not None
    else:
        import pkgutil
        return pkgutil.find_loader(importName) is not None


def assert_import_exists(importName):
    """Error if the given package or module name is not available for import.

    :raises ModuleNotFoundError or ImportError: If no module with the given name
         is available for import, denpending on the Python version.
    """
    if not import_exists(importName):
        if MAJOR >= 3:
            Error = ModuleNotFoundError
        else:
            Error = ImportError
        raise Error("Python module '{}' not found.".format(importName))


__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements :class:`Simplex`."""

import operator
from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from ..constraints import SimplexConstraint
from .data import convert_and_refine_arguments, convert_operands
from .exp_affine import AffineExpression, Constant
from .expression import refine_operands, validate_prediction
from .set import Set

_API_START = api_start(globals())
# -------------------------------


class Simplex(Set):
    r"""A (truncated, symmetrized) real simplex.

    :Definition:

    Let :math:`r \in \mathbb{R}_{\geq 0}` the specified radius and
    :math:`n \in \mathbb{Z}_{\geq 1}` an arbitrary dimensionality.

    1.  Without truncation and symmetrization, this is the nonnegative simplex

        .. math::

            \{x \in \mathbb{R}^n_{\geq 0} \mid \sum_{i = 1}^n x_i \leq r\}.

        For :math:`r = 1`, this is the standard (unit) :math:`n`-simplex.

    2.  With truncation but without symmetrization, this is the nonnegative
        simplex intersected with the :math:`\infty`-norm unit ball

        .. math::

            \{
                x \in \mathbb{R}^n_{\geq 0}
            \mid
                \sum_{i = 1}^n x_i \leq r \land x \leq 1
            \}.

        For :math:`r \leq 1`, this equals case (1).

    3.  With symmetrization but without truncation, this is the :math:`1`-norm
        ball of radius :math:`r`

        .. math::

            \{x \in \mathbb{R}^n \mid \sum_{i = 1}^n |x_i| \leq r\}.

    4.  With both symmetrization and truncation, this is the convex polytope

        .. math::

            \{
                x \in \mathbb{R}
            \mid
                \sum_{i = 1}^n |x_i| \leq r \land 0 \leq x \leq 1
            \}.

        For :math:`r \leq 1`, this equals case (3).
    """

    @convert_and_refine_arguments("radius")
    def __init__(self, radius=Constant(1), truncated=False, symmetrized=False):
        """Construct a :class:`Simplex`.

        :param radius: The radius of the simplex.
        :type radius:
            float or ~picos.expressions.AffineExpression
        """
        if not isinstance(radius, AffineExpression):
            raise TypeError("A simplex' radius must be given as a real affine "
                "expression, not as {}.".format(type(radius).__name__))
        elif not radius.scalar:
            raise TypeError("A simplex' radius must be scalar, not of shape {}."
                .format(glyphs.shape(radius.shape)))

        if radius.constant and radius.value <= 1:
            truncated = False

        var  = glyphs.free_var_name(radius.string)
        unit = "Unit " if radius.is1 else ""
        if not truncated and not symmetrized:
            typeStr = "{}Simplex".format(unit)
            symbStr = glyphs.set(glyphs.sep(glyphs.ge(var, 0),
                glyphs.le(glyphs.sum(var), radius.string)))
        elif truncated and not symmetrized:
            typeStr = "Box-Truncated {}Simplex".format(unit)
            symbStr = glyphs.set(glyphs.sep(glyphs.le(0, glyphs.le(var, 1)),
                glyphs.le(glyphs.sum(var), radius.string)))
        elif not truncated and symmetrized:
            typeStr = "{}1-norm Ball".format(unit)
            symbStr = glyphs.set(glyphs.sep(var,
                glyphs.le(glyphs.sum(glyphs.abs(var)), radius.string)))
        else:  # truncated and symmetrized
            typeStr = "Box-Truncated {}1-norm Ball".format(unit)
            symbStr = glyphs.set(glyphs.sep(glyphs.le(-1, glyphs.le(var, 1)),
                glyphs.le(glyphs.sum(glyphs.abs(var)), radius.string)))

        self._radius      = radius
        self._truncated   = truncated
        self._symmetrized = symmetrized

        Set.__init__(self, typeStr, symbStr)

    @property
    def radius(self):
        """The radius of the simplex."""
        return self._radius

    @property
    def truncated(self):
        r"""Whether this is intersected with the unit :math:`\infty`-ball."""
        return self._truncated

    @property
    def symmetrized(self):
        """Wether the simplex is mirrored onto all orthants."""
        return self._symmetrized

    def _get_variables(self):
        return self._radius.variables

    def _replace_variables(self, var_map):
        return self.__class__(self._radius._replace_variables(var_map),
            self._truncated, self._symmetrized)

    Subtype = namedtuple("Subtype", ("truncated", "symmetrized"))

    def _get_subtype(self):
        return self.Subtype(self._truncated, self._symmetrized)

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        if relation == operator.__rshift__:
            if issubclass(other.clstype, AffineExpression):
                return SimplexConstraint.make_type(
                    argdim=other.subtype.dim,
                    truncated=subtype.truncated,
                    symmetrized=subtype.symmetrized)

        return NotImplemented

    @convert_operands()
    @validate_prediction
    @refine_operands()
    def __rshift__(self, element):
        if isinstance(element, AffineExpression):
            return SimplexConstraint(self, element)
        else:
            return NotImplemented


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2012-2017 Guillaume Sagnol
# Copyright (C)      2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Functions for writing optimization problems to a file."""

# ------------------------------------------------------------------------------
# TODO: This is a temporary solution using outdated code. Ideally, writing to
#       a file should be integrated with the reformulation pipeline e.g. by
#       distinguishing a SolutionStrategy and an ExportStrategy.
# ------------------------------------------------------------------------------

from itertools import chain

import cvxopt
import numpy

from ..apidoc import api_end, api_start
from ..constraints import (
    AffineConstraint,
    LMIConstraint,
    RSOCConstraint,
    SOCConstraint,
)

from ..solvers import CVXOPTSolver, CPLEXSolver, MOSEKSolver, GurobiSolver
from ..expressions import IntegerVariable, BinaryVariable, CONTINUOUS_VARTYPES
from ..expressions.vectorizations import FullVectorization

_API_START = api_start(globals())
# -------------------------------


INFINITY = 1e16  #: A number deemed too large to appear in practice.


def write(picos_problem, filename, writer="picos"):
    r"""Write an optimization problem to a file.

    :param ~picos.Problem P: The problem to write.

    :param str filename: Path and name of the output file. The export format
        is inferred from the file extension. Supported extensions and their
        associated format are:

        * ``'.cbf'`` -- Conic Benchmark Format.

          This format is suitable for optimization problems involving second
          order and/or semidefinite cone constraints. This is a standard
          choice for conic optimization problems. Visit the website of
          `The Conic Benchmark Library <http://cblib.zib.de/>`_ or read
          `A benchmark library for conic mixed-integer and continuous
          optimization
          <http://www.optimization-online.org/DB_HTML/2014/03/4301.html>`_
          by Henrik A. Friberg for more information.

        * ``'.lp'`` -- `LP format
          <http://docs.mosek.com/6.0/pyapi/node022.html>`_.

          This format handles only linear constraints, unless the writer
          ``'cplex'`` is used. In the latter case the extended `CPLEX LP format
          <https://www.ibm.com/support/knowledgecenter/SSSA5P_12.10.0/
          ilog.odms.cplex.help/CPLEX/FileFormats/topics/LP.html>`_
          is used instead.

        * ``'.mps'`` -- `MPS format
          <http://docs.mosek.com/6.0/pyapi/node021.html>`_.

          As the writer, you need to choose one of ``'cplex'``, ``'gurobi'``
          or ``'mosek'``.

        * ``'.opf'`` -- `OPF format
          <http://docs.mosek.com/6.0/pyapi/node023.html>`_.

          As the writer, you need to choose ``'mosek'``.

        * ``'.dat-s'`` -- `Sparse SDPA format
          <http://sdpa.indsys.chuo-u.ac.jp/sdpa/download.html#sdpa>`_.

          This format is suitable for semidefinite programs. Second order
          cone constraints are stored as semidefinite constraints on an
          *arrow shaped* matrix.

    :param str writer: The default ``'picos'`` denotes PICOS' internal
        writer, which can export to *LP*, *CBF*, and *Sparse SDPA* formats.
        If CPLEX, Gurobi or MOSEK is installed, you can choose ``'cplex'``,
        ``'gurobi'``, or ``'mosek'``, respectively, to make use of that
        solver's export function and get access to more formats.

    .. warning::

        For problems involving a symmetric matrix variable :math:`X`
        (typically, semidefinite programs), the expressions involving
        :math:`X` are stored in PICOS as a function of :math:`svec(X)`, the
        symmetric vectorized form of :math:`X` (see `Dattorro, ch.2.2.2.1
        <http://meboo.convexoptimization.com/Meboo.html>`_),
        and are also exported in that form. As a result, using an external
        solver on a problem description file exported by PICOS will also
        yield a solution in this symmetric vectorized form.

        The CBF writer tries to write symmetric variables :math:`X` in the
        section ``PSDVAR`` of the .cbf file. However, this is possible only
        if the constraint :math:`X \succeq 0` appears in the problem, and no
        other LMI involves :math:`X`. If these two conditions are not
        satisfied, then the symmetric vectorization of :math:`X` is used as
        a (free) variable of the section ``VAR`` in the .cbf file, as
        explained in the previous paragraph.

    .. warning::

        This function is severly outdated and may fail or not function as
        advertised.
    """
    # HACK: This method abuses the internal problem representation of CVXOPT.
    # TODO: Add a proper method that transforms the problem into the canonical
    #       form required here, and, if applicable, make also CVXOPT use it.

    from .strategy import NoStrategyFound

    # We first prepare the problem so it becomes exportable
    COMMERCIAL_WRITERS = ['cplex', 'mosek', 'gurobi']
    if writer in COMMERCIAL_WRITERS:
        P = picos_problem.prepared(solver=writer)
    else:
        if picos_problem.continuous:
            # This should ensure that all quadratics have been cast as SOC
            # constraints.
            P = picos_problem.prepared(solver='cvxopt', assume_conic=True)
        else:
            try:
                P = picos_problem.prepared(solver=None)
            except NoStrategyFound:
                # FIXME: This will typically happen for integer linear programs
                # if no IP solver is available.
                raise NotImplementedError("You try to export a problem for "
                    "which no solver is available")

            assert 'Linear' in P.type, "only LINEAR integer problems can be " \
                "exported with the default writer"

    # TODO: Need proper way to detect exponential cone constraints. This catches
    #       only the Geometric Programs.
    if P.numberLSEConstraints:
        raise NotImplementedError("It is not possible (yet) to export GPs or "
            "problems with Exponential Cone Constraints")

    # automatic extension recognition
    if not (any(filename.endswith(ext) for ext in
                (".mps", ".opf", ".cbf", ".ptf", ".lp", ".dat-s"))):
        if writer == "gurobi":
            if (P.numberConeConstraints + P.numberQuadConstraints) == 0:
                filename += ".lp"
            else:
                filename += ".mps"
        elif writer == "cplex":
            filename += ".lp"
        elif writer == "mosek":
            if (P.numberConeConstraints + P.numberQuadConstraints
                    + P.numberSDPConstraints) == 0:
                filename += ".lp"
            elif P.numberQuadConstraints == 0:
                filename += ".cbf"
            else:
                filename += ".mps"
        elif writer == "picos":
            assert not P.numberQuadConstraints, \
                "Expected no quadratic constraints after call to 'prepared'."
            if (P.numberConeConstraints + P.numberSDPConstraints) == 0:
                filename += ".lp"
            elif P.numberConeConstraints == 0:
                filename += ".dat-s"
            else:
                filename += ".cbf"
        else:
            raise Exception("unexpected writer")

    if writer == "cplex":
        cpl = CPLEXSolver(P)
        cpl._load_problem()
        cpl.int.write(filename)

    elif writer == "mosek":
        if filename.endswith(".cbf"):
            # This ensures that all quadratics are converted to SOC
            P = picos_problem.prepared(solver='cvxopt')
        msk = MOSEKSolver(P)
        msk._load_problem()
        msk.int.writedata(filename)

    elif writer == "gurobi":
        grb = GurobiSolver(P)
        grb._load_problem()
        grb.int.write(filename)

    elif writer == "picos":
        if filename[-3:] == ".lp":
            _write_lp(P, filename)
        elif filename[-6:] == ".dat-s":
            _write_sdpa(P, filename)
        elif filename[-4:] == ".cbf":
            _write_cbf(P, filename)
        else:
            raise Exception("unexpected file extension")
    else:
        raise Exception("unknown writer")


def _write_lp(P, filename):
    """Write the problem to a file in LP format."""
    # add extension
    if filename[-3:] != ".lp":
        filename += ".lp"
    # check lp compatibility
    if (
            P.numberConeConstraints
            + P.numberQuadConstraints
            + P.numberLSEConstraints
            + P.numberSDPConstraints
    ) > 0:
        raise Exception("the picos LP writer only accepts (MI)LP")
    # open file
    f = open(filename, "w")
    f.write("\\* file " + filename + " generated by picos*\\\n")

    # HACK: This abuses the internal problem representation of CVXOPT.
    if P.continuous:
        localCvxoptInstance = CVXOPTSolver(P)
    else:
        localCvxoptInstance = CVXOPTSolver(P.continuous_relaxation())
    localCvxoptInstance.import_variable_bounds = False
    localCvxoptInstance._load_problem()
    cvxoptVars = localCvxoptInstance.internal_problem()

    # variable names
    varnames = {}
    i = 0
    for name, v in P.variables.items():
        # full vectorization is used, name variables as x, x[j] or x[j,k]
        if isinstance(v._vec, FullVectorization):
            for ind in range(v.dim):
                if v.size == (1, 1):
                    varnames[i] = name
                elif v.size[1] == 1:
                    varnames[i] = name + "(" + str(ind) + ")"
                else:
                    k, j = divmod(ind, v.size[0])
                    varnames[i] = name + "(" + str(j) + "," + str(k) + ")"
                varnames[i] = varnames[i].replace("[", "(")
                varnames[i] = varnames[i].replace("]", ")")
                i += 1
        else:
            for ind in range(v.dim):
                varnames[i] = 'vec_' + name + "(" + str(ind) + ")"
                i += 1

    # affexpr writer
    def affexp_writer(constraint_name, indices, coefs):
        s = ""
        s += constraint_name
        s += " : "
        start = True
        for (i, v) in zip(indices, coefs):
            if v > 0 and not (start):
                s += "+ "
            s += "%.12g" % v
            s += " "
            s += varnames[i]
            # not the first term anymore
            start = False
        if not (coefs):
            s += "0.0 "
            s += varnames[0]
        return s

    print("writing problem in " + filename + "...")

    # objective
    if P.objective.direction == "max":
        f.write("Maximize\n")
        # max handled directly
        cvxoptVars["c"] = -cvxoptVars["c"]
    else:
        f.write("Minimize\n")
    I = cvxopt.sparse(cvxoptVars["c"]).I
    V = cvxopt.sparse(cvxoptVars["c"]).V

    f.write(affexp_writer("obj", I, V))
    f.write("\n")

    f.write("Subject To\n")
    bounds = {}
    # equality constraints:
    Ai, Aj, Av = (cvxoptVars["A"].I, cvxoptVars["A"].J, cvxoptVars["A"].V)
    ijvs = sorted(zip(Ai, Aj, Av))
    del Ai, Aj, Av
    itojv = {}
    lasti = -1
    for (i, j, v) in ijvs:
        if i == lasti:
            itojv[i].append((j, v))
        else:
            lasti = i
            itojv[i] = [(j, v)]
    ieq = 0
    for i, jv in itojv.items():
        J = [jvk[0] for jvk in jv]
        V = [jvk[1] for jvk in jv]
        if len(J) == 1:
            # fixed variable
            b = cvxoptVars["b"][i] / V[0]
            bounds[J[0]] = (b, b)
        else:
            # affine equality
            b = cvxoptVars["b"][i]
            f.write(affexp_writer("eq" + str(ieq), J, V))
            f.write(" = ")
            f.write("%.12g" % b)
            f.write("\n")
            ieq += 1

    # inequality constraints:
    Gli, Glj, Glv = (cvxoptVars["Gl"].I, cvxoptVars["Gl"].J, cvxoptVars["Gl"].V)
    ijvs = sorted(zip(Gli, Glj, Glv))
    del Gli, Glj, Glv
    itojv = {}
    lasti = -1
    for (i, j, v) in ijvs:
        if i == lasti:
            itojv[i].append((j, v))
        else:
            lasti = i
            itojv[i] = [(j, v)]
    iaff = 0
    for i, jv in itojv.items():
        J = [jvk[0] for jvk in jv]
        V = [jvk[1] for jvk in jv]
        b = cvxoptVars["hl"][i]
        f.write(affexp_writer("in" + str(iaff), J, V))
        f.write(" <= ")
        f.write("%.12g" % b)
        f.write("\n")
        iaff += 1

    # variable bounds
    # retrieve as a dictionary {index -> (lo,up)}
    i_var = 0
    for var in P.variables.values():
        for ind, lo in var.bound_dicts[0].items():
            (current_lo, current_up) = bounds.get(
                i_var + ind, (-INFINITY, INFINITY))
            new_lo = max(lo, current_lo)
            bounds[i_var + ind] = (new_lo, current_up)
        for ind, up in var.bound_dicts[1].items():
            (current_lo, current_up) = bounds.get(
                i_var + ind, (-INFINITY, INFINITY))
            new_up = min(up, current_up)
            bounds[i_var + ind] = (current_lo, new_up)
        i_var += var.dim

    f.write("Bounds\n")
    for i in range(P.numberOfVars):
        if i in bounds:
            bl, bu = bounds[i]
        else:
            bl, bu = -INFINITY, INFINITY
        if bl == -INFINITY and bu == INFINITY:
            f.write(varnames[i] + " free")
        elif bl == bu:
            f.write(varnames[i] + (" = %.12g" % bl))
        elif bl < bu:
            if bl == -INFINITY:
                f.write("-inf <= ")
            else:
                f.write("%.12g" % bl)
                f.write(" <= ")
            f.write(varnames[i])
            if bu == INFINITY:
                f.write("<= +inf")
            else:
                f.write(" <= ")
                f.write("%.12g" % bu)
        f.write("\n")

    # general integers
    f.write("Generals\n")
    i_var = 0
    for name, v in P.variables.items():
        if isinstance(v, IntegerVariable):
            for ind in range(v.dim):
                f.write(varnames[i_var + ind] + "\n")
        i_var += v.dim

    # binary variables
    f.write("Binaries\n")
    i_var = 0
    for name, v in P.variables.items():
        if isinstance(v, BinaryVariable):
            for ind in range(v.dim):
                f.write(varnames[i_var + ind] + "\n")
        i_var += v.dim

    f.write("End\n")
    print("done.")
    f.close()


def _write_sdpa(P, filename):
    """Write the problem to a file in Sparse SDPA format."""
    # HACK: This abuses the internal problem representation of CVXOPT.
    localCvxoptInstance = CVXOPTSolver(P)
    localCvxoptInstance._load_problem()
    cvxoptVars = localCvxoptInstance.internal_problem()

    dims = {}
    dims["s"] = [int(numpy.sqrt(Gsi.size[0])) for Gsi in cvxoptVars["Gs"]]
    dims["l"] = cvxoptVars["Gl"].size[0]
    dims["q"] = [Gqi.size[0] for Gqi in cvxoptVars["Gq"]]
    G = cvxoptVars["Gl"]
    h = cvxoptVars["hl"]

    # handle the equalities as 2 ineq
    if cvxoptVars["A"].size[0] > 0:
        G = cvxopt.sparse([G, cvxoptVars["A"]])
        G = cvxopt.sparse([G, -cvxoptVars["A"]])
        h = cvxopt.matrix([h, cvxoptVars["b"]])
        h = cvxopt.matrix([h, -cvxoptVars["b"]])
        dims["l"] += 2 * cvxoptVars["A"].size[0]

    for i in range(len(dims["q"])):
        G = cvxopt.sparse([G, cvxoptVars["Gq"][i]])
        h = cvxopt.matrix([h, cvxoptVars["hq"][i]])

    for i in range(len(dims["s"])):
        G = cvxopt.sparse([G, cvxoptVars["Gs"][i]])
        h = cvxopt.matrix([h, cvxoptVars["hs"][i]])

    # Remove the lines in A and b corresponding to 0==0
    JP = list(set(cvxoptVars["A"].I))
    IP = range(len(JP))
    VP = [1] * len(JP)

    # is there a constraint of the form 0==a(a not 0) ?
    if any([b for (i, b) in enumerate(cvxoptVars["b"]) if i not in JP]):
        raise Exception("infeasible constraint of the form 0=a")

    from cvxopt import sparse, spmatrix

    PP = spmatrix(VP, IP, JP, (len(IP), cvxoptVars["A"].size[0]))
    cvxoptVars["A"] = PP * cvxoptVars["A"]
    cvxoptVars["b"] = PP * cvxoptVars["b"]
    c = cvxoptVars["c"]
    # ------------------------------------------------------------#
    # make A,B,and blockstruct.                                  #
    # This code is a modification of the conelp function in SMCP #
    # ------------------------------------------------------------#
    Nl = dims["l"]
    Nq = dims["q"]
    Ns = dims["s"]
    if not Nl:
        Nl = 0

    P_m = G.size[1]

    P_b = -c
    P_blockstruct = []
    if Nl:
        P_blockstruct.append(-Nl)
    for i in Nq:
        P_blockstruct.append(i)
    for i in Ns:
        P_blockstruct.append(i)

    # write data
    # add extension
    if filename[-6:] != ".dat-s":
        filename += ".dat-s"

    # open file
    f = open(filename, "w")
    f.write('"file ' + filename + ' generated by picos"\n')
    if P.options.verbosity >= 1:
        print("writing problem in " + filename + "...")
    f.write(str(P.numberOfVars) + " = number of vars\n")
    f.write(str(len(P_blockstruct)) + " = number of blocs\n")
    # bloc structure
    f.write(str(P_blockstruct).replace("[", "(").replace("]", ")"))
    f.write(" = BlocStructure\n")
    # c vector (objective)
    f.write(str(list(-P_b)).replace("[", "{").replace("]", "}"))
    f.write("\n")
    # coefs
    for k in range(P_m + 1):
        if k != 0:
            v = sparse(G[:, k - 1])
        else:
            v = +sparse(h)

        ptr = 0
        block = 0
        # lin. constraints
        if Nl:
            u = v[:Nl]
            for i, j, value in zip(u.I, u.I, u.V):
                f.write(
                    "{0}\t{1}\t{2}\t{3}\t{4}\n".format(
                        k, block + 1, j + 1, i + 1, -value
                    )
                )
            ptr += Nl
            block += 1

        # SOC constraints
        for nq in Nq:
            u0 = v[ptr]
            u1 = v[ptr + 1:ptr + nq]
            tmp = spmatrix(
                u1.V, [nq - 1 for j in range(len(u1))], u1.I, (nq, nq)
            )
            if not u0 == 0.0:
                tmp += spmatrix(u0, range(nq), range(nq), (nq, nq))
            for i, j, value in zip(tmp.I, tmp.J, tmp.V):
                f.write(
                    "{0}\t{1}\t{2}\t{3}\t{4}\n".format(
                        k, block + 1, j + 1, i + 1, -value
                    )
                )
            ptr += nq
            block += 1

        # SDP constraints
        for ns in Ns:
            u = v[ptr:ptr + ns ** 2]
            for index_k, index in enumerate(u.I):
                j, i = divmod(index, ns)
                if j <= i:
                    f.write(
                        "{0}\t{1}\t{2}\t{3}\t{4}\n".format(
                            k, block + 1, j + 1, i + 1, -u.V[index_k]
                        )
                    )
            ptr += ns ** 2
            block += 1

    f.close()


# This is an ugly function from a former version of picos
# It separates a linear constraint between 'plain' vars and matrix 'bar'
# variables J and V denote the sparse indices/values of the constraints for the
# whole (s-)vectorized vector (without offset)
def _separate_linear_cons_plain_bar_vars(J, V, idx_sdp_vars):
    # sparse values of the constraint for 'plain' variables
    jj = []
    vv = []
    # sparse values of the constraint for the next svec bar variable
    js = []
    vs = []
    mats = []
    offset = 0
    if idx_sdp_vars:
        idxsdpvars = [ti for ti in idx_sdp_vars]
        nextsdp = idxsdpvars.pop()
    else:
        return J, V, []
    for (j, v) in zip(J, V):
        if j < nextsdp[0]:
            jj.append(j - offset)
            vv.append(v)
        elif j < nextsdp[1]:
            js.append(j - nextsdp[0])
            vs.append(v)
        else:
            while j >= nextsdp[1]:
                mats.append(
                    devectorize(
                        cvxopt.spmatrix(
                            vs,
                            js,
                            [0] * len(js),
                            (nextsdp[1] - nextsdp[0], 1)
                        )).T)
                js = []
                vs = []
                offset += (nextsdp[1] - nextsdp[0])
                try:
                    nextsdp = idxsdpvars.pop()
                except IndexError:
                    nextsdp = (float('inf'), float('inf'))
            if j < nextsdp[0]:
                jj.append(j - offset)
                vv.append(v)
            elif j < nextsdp[1]:
                js.append(j - nextsdp[0])
                vs.append(v)
    while len(mats) < len(idx_sdp_vars):
        mats.append(
            devectorize(
                cvxopt.spmatrix(
                    vs,
                    js,
                    [0] * len(js),
                    (nextsdp[1] - nextsdp[0], 1)
                )).T)
        js = []
        vs = []
        nextsdp = (0, 1)  # doesnt matter, it will be an empt matrix anyway
    return jj, vv, mats


def devectorize(vec):
    """Create a matrix from a symmetric vectorization."""
    from ..expressions.vectorizations import SymmetricVectorization
    v = vec.size[0]
    n = int((1 + 8 * v) ** 0.5 - 1) // 2
    return SymmetricVectorization((n, n)).devectorize(vec)


def _write_cbf(P, filename, uptri=False):
    """Write the problem to a file in Sparse SDPA format.

    :param bool uptri: Whether upper triangular elements of symmetric
        matrices are specified.
    """
    # write data
    # add extension
    if filename[-4:] != ".cbf":
        filename += ".cbf"

    # parse variables
    semidef_vars = set()
    for cons in P.constraints:
        cs = P.constraints[cons]
        if isinstance(cs, LMIConstraint):
            if cs.semidefVar:
                semidef_vars.add(cs.semidefVar)

    NUMVAR_SCALAR = int(
        sum(
            [var.dim
             for var in P.variables.values()
             if var not in semidef_vars]
        )
    )

    ind = 0
    indices = []
    start_indices = {}
    for v in P.variables.values():
        indices.append((ind, ind + v.dim, v))
        start_indices[v] = ind
        ind += v.dim

    indices = sorted(indices)
    idxsdpvars = [
        (si, ei) for (si, ei, v) in indices[::-1] if v in semidef_vars]
    # search if some semidef vars are implied in other semidef constraints
    PSD_not_handled = []
    for c in P.constraints.values():
        if isinstance(c, LMIConstraint) and c not in semidef_vars:
            for v in (c.lhs - c.rhs)._coefs:
                if v in semidef_vars:
                    start_index = start_indices[v]
                    idx = (start_index, start_index + v.dim)
                    if idx in idxsdpvars:
                        PSD_not_handled.append(v)
                        NUMVAR_SCALAR += idx[1] - idx[0]
                        idxsdpvars.remove(idx)

    barvars = bool(idxsdpvars)

    # find integer variables, retrieve bounds and put 0-1 bounds on binaries
    ints = []
    ind = 0
    varbounds_lo = {}
    varbounds_up = {}

    for k, var in P.variables.items():
        if isinstance(var, BinaryVariable):
            for relind, absind in enumerate(range(ind, ind + var.dim)):
                ints.append(absind)
                clb = var.bound_dicts[0].get(relind, -INFINITY)
                cub = var.bound_dicts[1].get(relind, INFINITY)
                varbounds_lo[absind] = max(0.0, clb)
                varbounds_up[absind] = min(1.0, cub)

        elif (isinstance(var, IntegerVariable) or
              isinstance(var, CONTINUOUS_VARTYPES)):
            for relind, absind in enumerate(range(ind, ind + var.dim)):
                if isinstance(var, IntegerVariable):
                    ints.append(absind)
                clb = var.bound_dicts[0].get(relind, -INFINITY)
                cub = var.bound_dicts[1].get(relind, INFINITY)
                varbounds_lo[absind] = clb
                varbounds_up[absind] = cub

        else:
            raise Exception("variable type not handled by _write_cbf()")
        ind += var.dim

    if barvars:
        ints, _, mats = _separate_linear_cons_plain_bar_vars(
            ints, [0.0] * len(ints), idxsdpvars
        )
        if any([bool(mat) for mat in mats]):
            raise Exception(
                "semidef vars with integer elements are not supported"
            )

    # open file
    f = open(filename, "w")
    f.write("#file " + filename + " generated by picos\n")
    print("writing problem in " + filename + "...")

    f.write("VER\n")
    f.write("1\n\n")

    f.write("OBJSENSE\n")
    if P.objective.direction == "max":
        f.write("MAX\n\n")
    else:
        f.write("MIN\n\n")

    # VARIABLEs

    if barvars:
        f.write("PSDVAR\n")
        f.write(str(len(idxsdpvars)) + "\n")
        for si, ei in idxsdpvars:
            ni = int(((8 * (ei - si) + 1) ** 0.5 - 1) / 2.0)
            f.write(str(ni) + "\n")
        f.write("\n")

    # bounds
    cones = []
    conecons = []
    Acoord = []
    Bcoord = []
    iaff = 0
    offset = 0
    for si, ei, v in indices:
        if v in semidef_vars and not (v in PSD_not_handled):
            offset += ei - si
        else:
            if all(varbounds_lo[ind] == 0 for ind in range(si, ei)):
                cones.append(("L+", ei - si))
            elif all(varbounds_up[ind] == 0 for ind in range(si, ei)):
                cones.append(("L-", ei - si))
            else:
                cones.append(("F", ei - si))
            if any(varbounds_lo[ind] != 0 for ind in range(si, ei)):
                for ind in range(si, ei):
                    l = varbounds_lo[ind]
                    if l - 1 > -INFINITY:
                        Acoord.append((iaff, ind - offset, 1.0))
                        Bcoord.append((iaff, -l))
                        iaff += 1
            if any(varbounds_up[ind] != 0 for ind in range(si, ei)):
                for ind in range(si, ei):
                    u = varbounds_up[ind]
                    if u + 1 < INFINITY:
                        Acoord.append((iaff, ind - offset, -1.0))
                        Bcoord.append((iaff, u))
                        iaff += 1
    if iaff:
        conecons.append(("L+", iaff))

    f.write("VAR\n")
    f.write(str(NUMVAR_SCALAR) + " " + str(len(cones)) + "\n")
    for tp, n in cones:
        f.write(tp + " " + str(n) + "\n")

    f.write("\n")

    # integers
    if ints:
        f.write("INT\n")
        f.write(str(len(ints)) + "\n")
        for i in ints:
            f.write(str(i) + "\n")
        f.write("\n")

    # constraints
    psdcons = []
    isdp = 0
    Fcoord = []
    Hcoord = []
    Dcoord = []
    ObjAcoord = []
    ObjBcoord = []
    ObjFcoord = []

    # dummy constraint for the objective
    dummy_cons = P.objective.function >= 0
    setattr(dummy_cons, "dummycon", None)

    for cons in chain((dummy_cons,), P.constraints.values()):
        if isinstance(cons, LMIConstraint):
            v = cons.semidefVar
            if v is not None and v not in PSD_not_handled:
                continue

        # get sparse indices
        if isinstance(cons, AffineConstraint):
            expcone = cons.lhs - cons.rhs
            if hasattr(cons, "dummycon"):
                conetype = "0"  # Dummy type for the objective function.
            elif cons.is_equality():
                conetype = "L="
            elif cons.is_increasing():
                conetype = "L-"
            elif cons.is_decreasing():
                conetype = "L+"
            else:
                assert False, "Unexpected constraint relation."
        elif isinstance(cons, SOCConstraint):
            expcone = (cons.ub) // (cons.ne[:])
            conetype = "Q"
        elif isinstance(cons, RSOCConstraint):
            expcone = (cons.ub1) // (0.5 * cons.ub2) // (cons.ne[:])
            conetype = "QR"
        elif isinstance(cons, LMIConstraint):
            if cons.is_increasing():
                expcone = cons.rhs - cons.lhs
                conetype = None
            elif cons.is_decreasing():
                expcone = cons.lhs - cons.rhs
                conetype = None
            else:
                assert False, "Unexpected constraint relation."
        else:
            assert False, "Unexpected constraint type."

        ijv = []
        for var, fact in expcone._coefs.items():
            if not isinstance(fact, cvxopt.base.spmatrix):
                fact = cvxopt.sparse(fact)
            sj = start_indices[var]
            ijv.extend(zip(fact.I, fact.J + sj, fact.V))
        ijvs = sorted(ijv)

        itojv = {}
        lasti = -1
        for (i, j, v) in ijvs:
            if i == lasti:
                itojv[i].append((j, v))
            else:
                lasti = i
                itojv[i] = [(j, v)]

        if conetype:
            if conetype != "0":
                dim = expcone.size[0] * expcone.size[1]
                conecons.append((conetype, dim))
        else:
            dim = expcone.size[0]
            psdcons.append(dim)

        if conetype:
            for i, jv in itojv.items():
                J = [jvk[0] for jvk in jv]
                V = [jvk[1] for jvk in jv]
                J, V, mats = _separate_linear_cons_plain_bar_vars(
                    J, V, idxsdpvars)
                for j, v in zip(J, V):
                    if conetype != "0":
                        Acoord.append((iaff + i, j, v))
                    else:
                        ObjAcoord.append((j, v))
                for k, mat in enumerate(mats):
                    for row, col, v in zip(mat.I, mat.J, mat.V):
                        if conetype != "0":
                            Fcoord.append((iaff + i, k, row, col, v))
                        else:
                            ObjFcoord.append((k, row, col, v))
                        if uptri and row != col:
                            if conetype != "0":
                                Fcoord.append((iaff + i, k, col, row, v))
                            else:
                                ObjFcoord.append((k, col, row, v))
            constant = expcone._const
            if not (constant is None):
                constant = cvxopt.sparse(constant)
                for i, v in zip(constant.I, constant.V):
                    if conetype != "0":
                        Bcoord.append((iaff + i, v))
                    else:
                        ObjBcoord.append(v)
        else:
            for i, jv in itojv.items():
                col, row = divmod(i, dim)
                if not (uptri) and row < col:
                    continue
                J = [jvk[0] for jvk in jv]
                V = [jvk[1] for jvk in jv]
                J, V, mats = _separate_linear_cons_plain_bar_vars(
                    J, V, idxsdpvars)
                if any([bool(m) for m in mats]):
                    raise Exception("SDP cons should not depend on PSD var")
                for j, v in zip(J, V):
                    Hcoord.append((isdp, j, row, col, v))

            constant = expcone._const
            if not (constant is None):
                constant = cvxopt.sparse(constant)
                for i, v in zip(constant.I, constant.V):
                    col, row = divmod(i, dim)
                    if row < col:
                        continue
                    Dcoord.append((isdp, row, col, v))

        if conetype:
            if conetype != "0":
                iaff += dim
        else:
            isdp += 1

    if iaff > 0:
        f.write("CON\n")
        f.write(str(iaff) + " " + str(len(conecons)) + "\n")
        for tp, n in conecons:
            f.write(tp + " " + str(n))
            f.write("\n")

        f.write("\n")

    if isdp > 0:
        f.write("PSDCON\n")
        f.write(str(isdp) + "\n")
        for n in psdcons:
            f.write(str(n) + "\n")
        f.write("\n")

    if ObjFcoord:
        f.write("OBJFCOORD\n")
        f.write(str(len(ObjFcoord)) + "\n")
        for (k, row, col, v) in ObjFcoord:
            f.write("{0} {1} {2} {3}\n".format(k, row, col, v))
        f.write("\n")

    if ObjAcoord:
        f.write("OBJACOORD\n")
        f.write(str(len(ObjAcoord)) + "\n")
        for (j, v) in ObjAcoord:
            f.write("{0} {1}\n".format(j, v))
        f.write("\n")

    if ObjBcoord:
        f.write("OBJBCOORD\n")
        v = ObjBcoord[0]
        f.write("{0}\n".format(v))
        f.write("\n")

    if Fcoord:
        f.write("FCOORD\n")
        f.write(str(len(Fcoord)) + "\n")
        for (i, k, row, col, v) in Fcoord:
            f.write("{0} {1} {2} {3} {4}\n".format(i, k, row, col, v))
        f.write("\n")

    if Acoord:
        f.write("ACOORD\n")
        f.write(str(len(Acoord)) + "\n")
        for (i, j, v) in Acoord:
            f.write("{0} {1} {2}\n".format(i, j, v))
        f.write("\n")

    if Bcoord:
        f.write("BCOORD\n")
        f.write(str(len(Bcoord)) + "\n")
        for (i, v) in Bcoord:
            f.write("{0} {1}\n".format(i, v))
        f.write("\n")

    if Hcoord:
        f.write("HCOORD\n")
        f.write(str(len(Hcoord)) + "\n")
        for (i, j, row, col, v) in Hcoord:
            f.write("{0} {1} {2} {3} {4}\n".format(i, j, row, col, v))
        f.write("\n")

    if Dcoord:
        f.write("DCOORD\n")
        f.write(str(len(Dcoord)) + "\n")
        for (i, row, col, v) in Dcoord:
            f.write("{0} {1} {2} {3}\n".format(i, row, col, v))
        f.write("\n")

    print("done.")
    f.close()


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2017-2018 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Optimization solver interfaces.

This package contains the interfaces to the optimization solvers that PICOS uses
as its backend. You do not need to instanciate any of the solver classes
directly; if you want to select a particular solver, it is most convenient to
supply it to :func:`Problem.solve <picos.Problem.solve>` via the ``solver``
keyword argument.
"""

from ..apidoc import api_end, api_start
from ..legacy import throw_deprecation_warning


_API_START = api_start(globals())
# -------------------------------

# Import the solver base class and exceptions.
from .solver import (Solver, SolverError, ProblemUpdateError,  # noqa
    OptionError, UnsupportedOptionError,
    ConflictingOptionsError, DependentOptionError, OptionValueError)

# Import all solvers.
from .solver_cplex  import CPLEXSolver  # noqa
from .solver_cvxopt import CVXOPTSolver  # noqa
from .solver_ecos   import ECOSSolver  # noqa
from .solver_glpk   import GLPKSolver  # noqa
from .solver_gurobi import GurobiSolver  # noqa
from .solver_mosek  import MOSEKSolver  # noqa
from .solver_mskfsn import MOSEKFusionSolver  # noqa
from .solver_scip   import SCIPSolver  # noqa
from .solver_smcp   import SMCPSolver  # noqa


# Map solver names to their implementation classes.
_solvers = {solver.names()[0]: solver for solver in (
    CPLEXSolver,
    CVXOPTSolver,
    ECOSSolver,
    GLPKSolver,
    GurobiSolver,
    MOSEKSolver,
    MOSEKFusionSolver,
    SCIPSolver,
    SMCPSolver
)}

# Make sure all solvers inherit from their abstract base class.
assert all(issubclass(solver, Solver) for solver in _solvers.values())


def get_solver(name):
    """Return the implementation class of the solver with the given name."""
    return _solvers[name]


def get_solver_name(solver):
    """Return the registry name of a solver instance."""
    for name, solverClass in _solvers.items():
        if isinstance(solver, solverClass):
            return name
    raise LookupError("The given object's type is not regisered as a solver.")


def all_solvers():
    """Return a dictionary mapping solver names to implementation classes."""
    return _solvers.copy()


def available_solvers(problem=None):
    """Return a sorted list of names of available solvers.

    :param problem: DEPRECATED
    """
    if problem is not None:
        throw_deprecation_warning(
            "Arguments 'problem' to picos.available_solvers is deprecated.")

    return sorted(
        name for name, solver in _solvers.items() if solver.available())


# --------------------------------------
__all__ = api_end(_API_START, globals())
